#ifndef _WAR_LIST_
#define _WAR_LIST_

#include <Locker.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// A list class that uses an array, it is currently NOT thread safe.

template<class Item> class WarList
{
public:
	WarList();
	WarList(uint16 size,uint16 step_size);
	~WarList();

	bool GetNextItem(Item *item,uint16 *index);
	uint16 AddItem(Item item);
	bool AddItemAt(Item item,uint16 index);
	bool HasItem(Item item);

	bool RemoveItem(Item item);
	Item RemoveItemAt(uint16 index);
	uint16 CountItems();
	uint16 FirstEmpty();

	bool IsEmpty();
	void Empty();
	Item ItemAt(uint16 index);
	
	bool ExpandBy(uint16 num_items);
private:

	Item *m_list;
	int32 m_allocated_items;
	uint16 m_step;
	int32 m_last_item;
	uint16 m_num_items;
};

template <class Item>
WarList<Item>::WarList()
{
	m_allocated_items=10;
	m_list = new Item[m_allocated_items];
	memset(m_list,0,m_allocated_items*sizeof(Item));

	m_step=5;
	m_last_item=-1;
	m_num_items=0;
}

template <class Item>
WarList<Item>::WarList(uint16 size,uint16 step_size)
{
	m_allocated_items=size;
	m_list = new Item[m_allocated_items];
	memset(m_list,0,m_allocated_items*sizeof(Item));
	m_step=step_size;
	m_last_item=-1;
	m_num_items=0;
}

template <class Item>
WarList<Item>::~WarList()
{
	delete []m_list;
}

template <class Item>
bool WarList<Item>::GetNextItem(Item *item,uint16 *index)
{
	if (*index>=m_allocated_items || *index>m_last_item)
	{
		*index=0;
		return false;
	}

	// skip the empty entries
	while(*index < m_allocated_items && m_list[*index] == 0x0)
		*index+=1;

	// is this entry used? if not it must be the last entry
	if (*index <= m_last_item)
	{
		*item=m_list[*index];
		*index+=1;
		return true;
	}
	*index=0;
	return false;
}

template <class Item>
uint16 WarList<Item>::AddItem(Item item)
{
	uint16 index=0;

	while(index < m_allocated_items && m_list[index] != 0x0)
		index++;

	// the unit id number can't be greater than 65535
	if (index == 0xFFFF) return -1;

	// if we add at the end of the list
	if (index > m_last_item)
	{
		m_last_item++;
		// check that we haven't ran out of memory for the list
		if (index == m_allocated_items-1) ExpandBy(m_step);
	}
	m_num_items++;
	m_list[index] = item;
	return index;
}

template <class Item>
bool WarList<Item>::AddItemAt(Item item,uint16 index)
{
	// check that we haven't ran out of memory for the list
	if (index >= m_allocated_items)
		if (!ExpandBy(index - m_allocated_items)) return false;

	if (m_list[index] != NULL) return false;
	if (index > m_last_item) m_last_item = index;
	m_list[index] = item;
	m_num_items++;
	return true;
}

template <class Item>
bool WarList<Item>::HasItem(Item item)
{
	uint16 index = 0;
	while(index <= m_last_item && m_list[index] != item)
		index++;
	if (index > m_last_item) return false;
	return true;
}

template <class Item>
uint16 WarList<Item>::FirstEmpty()
{
	uint16 index = 0;
	while(index < m_allocated_items && m_list[index])
		index++;
	return index;
}

template <class Item>
bool WarList<Item>::RemoveItem(Item item)
{
	uint16 index = 0;
	while(index <= m_last_item && m_list[index] != item)
		index++;
	if (index > m_last_item) return false;

	if (index == m_last_item)
		m_last_item--;
	
	m_num_items--;
	m_list[index] = 0;
	return true;
}

template <class Item>
Item WarList<Item>::RemoveItemAt(uint16 index)
{
	Item ret;
	if (index > m_last_item) return NULL;
	ret = m_list[index];

	if (index == m_last_item)
	{
		if (index>0)
		{
			int32 index2 = index-1;
			while (index2 >= 0 && m_list[index2] == 0)
				index2--;
			m_last_item = index2;
		}
		else m_last_item = -1;
	}

	m_list[index] = 0;
	if (ret) m_num_items--;
	return ret;
}

// expands the list of allocated items by the given number of elements
template <class Item>
bool WarList<Item>::ExpandBy(uint16 item_count)
{
	if (m_allocated_items + item_count >= 0xffff) return false;

	Item *new_list = new Item[m_allocated_items + item_count];
	uint16 i=0;

	for (;i<m_allocated_items;i++)
		new_list[i] = m_list[i];
	
	for (;i<m_allocated_items + item_count;i++)
		new_list[i] = 0;

	m_allocated_items += item_count;
	delete []m_list;
	m_list = new_list;

	return true;
}

// inline definitions

template <class Item>
inline Item WarList<Item>::ItemAt(uint16 index)
{
	if (index > m_last_item) return NULL;
	return m_list[index];
}

template <class Item>
inline uint16 WarList<Item>::CountItems()
{
	return m_num_items;
}

template <class Item>
inline bool WarList<Item>::IsEmpty()
{
	return (m_last_item == -1);
}

template <class Item>
inline void WarList<Item>::Empty()
{
	m_last_item = -1;
	m_num_items = 0;
}
#endif // #ifndef _WAR_LIST_
