#ifndef _WAR_MAP_
#define _WAR_MAP_

#include <Locker.h>
#include <stdio.h>
#include <gl.h>

#include "WarTileset.h"
#include "WarVertex.h"
#include "WarSplineGrid.h"
#include "WarGrid.h"
#include "WarBlockMap.h"
#include "WarPoint.h"
#include "Global.h"

enum block_def // the numbers in the block info array
{
	WAR_GRASS=0,
	WAR_ROAD,
	WAR_SWAMP,
	WAR_OCEAN,
	WAR_RUINS,
	WAR_ROCKS,
	WAR_RIVER,
	WAR_CUSTOM1,
	WAR_CUSTOM2,
	WAR_CUSTOM3,
	WAR_CUSTOM4,
	WAR_CUSTOM5,
	WAR_CUSTOM6,
	WAR_CUSTOM7,
	WAR_CUSTOM8,
	WAR_CUSTOM9
};

class WarMap:public BLocker
{
public:
	WarMap(uint8 size,BDirectory *env_dir,char *environment);
	WarMap(BDirectory *map_dir,char *filename,BDirectory *env_dir);
	~WarMap();

	void Save(BDirectory *dir,char *filename);

	void GenerateMap();
	
	uint8 Size(); // returns the size constant

// –––––––––––––––––––––––––––––

	uint16 Width(); // returns the size of the map (in 16x16 blocks)
	uint16 Height();
	uint8 WidthShift(); // instead of multiplying with the Width() shift left with this value
	uint8 HeightShift();

// –––––––––––––––––––––––––––––

	uint32 BlockCount(); // the number of blocks on the map

// –––––––––––––––––––––––––––––

	// returns the flags set at the given offset of the map
	uint8 GetFlags(uint32 offs);
	// returns the flags set at the given point of the map
	uint8 GetFlags(uint16 x,uint16 y);
	// returns the flags set anywhere within the given rectangle of the map
	uint8 GetFlagsRect(int32 x,int32 y,uint16 width,uint16 height);

	// sets the given flags at the given offset
	void SetFlags(uint32 offs,uint8 flags);
	// clears the given flags on the whole map
	void ClearFlags(uint8 flags); // clear the specified flags for all the block

// –––––––––––––––––––––––––––––

	uint16 TileIndex(int32 x,int32 y);
	void SetTileIndex(int32 x,int32 y,uint16 index);

// –––––––––––––––––––––––––––––

	float GetHeight(int32 x,int32 y); // returns the height of the map at a given integer point
	float GetHeight(float x,float y); // returns the interpolated height of the map at a given point

	WarPoint Slope(int32 x,int32 y); // returns the slope att the given integer point
	WarPoint Slope(float x,float y); // returns the interpolated slope att the given point

// –––––––––––––––––––––––––––––

	uint8 Bump(int32 x,int32 y);
	void RecalcBump(int32 x,int32 y,bool dir,int32 level);

	WarTileset *Tileset(); // returns a pointer to the tilset used on this map

	uint8 Terrain(uint32 offs); // returns the terrain constant for the specified block
	
	float BoxElevation(int32 x,int32 y);
	float BoxHeight(int32 x,int32 y);

private:

	void init(uint8 size,BDirectory *env_dir,char *environment);

	uint32 m_block_count; // if you and the offset with this value, you never fall out of boundry
	uint8 m_size; // the size#

	uint8 *m_map_flags;	// size: width * height

	WarGrid<float> *m_height_map;
	WarGrid<WarPoint> *m_map_slope;
	
	WarBlockMap<uint16> *m_map_textures;

	int32 m_width,m_height; // the map size in blocks

	uint8 m_block_data[2048]; // the block info array
	uint8 m_width_shift,m_height_shift; // the map size shifts

	WarTileset *m_tiles;

	uint8 m_blocks_per_control_point_x;
	uint8 m_blocks_per_control_point_y;

	WarBlockMap<uint8> *m_box_height;
	WarBlockMap<float> *m_box_elevation;

	uint8 *bump_map;	// describes the "bumpyness" of the terrain
};

// inline definitions

inline float WarMap::BoxElevation(int32 x,int32 y)
{
	return m_box_elevation->Value(x,y);
}

inline float WarMap::BoxHeight(int32 x,int32 y)
{
	return m_box_height->Value(x,y);
}

inline WarTileset *WarMap::Tileset()
{
	return m_tiles;
}

inline uint8 WarMap::Size()
{
	return m_size;
}

inline uint16 WarMap::Width()
{
	return m_width;
}

inline uint16 WarMap::Height()
{
	return m_height;
}

inline uint8 WarMap::WidthShift()
{
	return m_width_shift;
}

inline uint8 WarMap::HeightShift()
{
	return m_height_shift;
}

// BLOCK

inline uint16 WarMap::TileIndex(int32 x,int32 y)
{
	return m_map_textures->Value(x,y);
}

inline void WarMap::SetTileIndex(int32 x,int32 y,uint16 index)
{
	m_map_textures->SetValue(x,y,index);
}

// FLAGS

inline uint8 WarMap::GetFlags(uint32 offs)
{
	if (offs <= m_block_count) return m_map_flags[offs];
	return 0;
}

inline uint8 WarMap::GetFlags(uint16 x,uint16 y)
{
	return GetFlags((y<<WidthShift())+x);
}

inline void WarMap::SetFlags(uint32 offs,uint8 flags)
{
	if (offs <= m_block_count) m_map_flags[offs] = flags;
}

inline uint32 WarMap::BlockCount()
{
	return m_block_count+1;
}

inline float WarMap::GetHeight(int32 fx,int32 fy)
{
	return m_height_map->Value(fx,fy);
}

inline float WarMap::GetHeight(float x,float y)
{
	int32 ix = (int32)x;
	int32 iy = (int32)y;

	float th = x - (float)ix;
	float tv = y - (float)iy;

/*
		   v1
		+---+-------+
		|   :       |
	 v4+...o.......+v2
	   |   :(th,tv)|
		|   :       |
		|   :       |
		+---+-------+
	(0,0)   v3
*/

	if (th == 0.0 && tv == 0.0) return m_height_map->Value(ix,iy);
	else if (th == 1.0 && tv == 0.0) return m_height_map->Value(ix+1,iy);
	else if (th == 1.0 && tv == 1.0) return m_height_map->Value(ix+1,iy+1);
	else if (th == 0.0 && tv == 1.0) return m_height_map->Value(ix,iy+1);

	float v1 = m_height_map->Value(ix,iy);
	float v2 = m_height_map->Value(ix+1,iy);
	float v3 = m_height_map->Value(ix+1,iy+1);
	float v4 = m_height_map->Value(ix,iy+1);

	return (
				(v3*tv + v2*(1.0-tv))*th +
				(v4*tv + v1*(1.0-tv))*(1.0-th) +
				(v2*th + v1*(1.0-th))*(1.0-tv) +
				(v3*th + v4*(1.0-th))*tv
			)/2.0;
}

inline WarPoint WarMap::Slope(int32 fx,int32 fy)
{
	return m_map_slope->Value(fx,fy);
}

inline WarPoint WarMap::Slope(float x,float y)
{
	int32 ix = (int32)x;
	int32 iy = (int32)y;

	float th = x - (float)ix;
	float tv = y - (float)iy;

/*
		   v1
		+---+-------+
		|   :       |
	 v4+...o.......+v2
	   |   :(th,tv)|
		|   :       |
		|   :       |
		+---+-------+
	(0,0)   v3
*/

	WarPoint v1 = m_map_slope->Value(ix,iy);
	WarPoint v2 = m_map_slope->Value(ix+1,iy);
	WarPoint v3 = m_map_slope->Value(ix+1,iy+1);
	WarPoint v4 = m_map_slope->Value(ix,iy+1);

	if (th == 0.0 && tv == 0.0) return v1;
	else if (th == 1.0 && tv == 0.0) return v2;
	else if (th == 1.0 && tv == 1.0) return v3;
	else if (th == 0.0 && tv == 1.0) return v4;

	return (
				(v3*tv + v2*(1.0f-tv))*th +
				(v4*tv + v1*(1.0f-tv))*(1.0f-th) +
				(v2*th + v1*(1.0f-th))*(1.0f-tv) +
				(v3*th + v4*(1.0f-th))*tv
			)/2.0;
}

inline void WarMap::ClearFlags(uint8 flags)
{
	for (uint32 i=0;i <= m_block_count;i++)
		m_map_flags[i]&=~flags;
}

inline uint8 WarMap::Bump(int32 x,int32 y)
{
/*
	ASSERT(x >= 0);
	ASSERT(y >= 0);
	ASSERT(x <= Width());
	ASSERT(y <= Height());
*/
	return bump_map[y * (Width() + 1) + x];
}

#endif // #ifndef _WAR_MAP_