#ifndef _WAR_CAMERA_
#define _WAR_CAMERA_

#include "WarMap.h"
#include "WarVertex.h"
#include "WarObject.h"
#include "WarEntityList.h"
#include "WarMatrix.h"
#include "WarBox.h"

#include <Rect.h>
#include <gl.h>
#include <GraphicsCard.h>
#include <List.h>

class WarCamera : public WarMatrix
{
public:
	WarCamera();
	WarCamera(WarVertex position);
	virtual ~WarCamera();

	virtual WarVertex Position(); // returns the position in a vertex
	virtual float XPosition();
	virtual float YPosition();
	virtual float ZPosition();

	virtual WarVertex Direction(); // returns the x, y and z angle (degrees) in a vertex
	virtual float XDirection();
	virtual float YDirection();
	virtual float ZDirection();


	void Cycle(float frame_time,WarEntityList *war_entity_list,WarMap *map);
	void Swap();

	void Scroll(float horiz,float vert);

	void AddMovementSpeed(WarVertex delta_pos);
	void AddRotationSpeed(WarVertex delta_angles);
	void AddScrollSpeed(float vert,float horiz);

// zeroes the cameras rotation, movement and scroll speed
	void Stop();

// makes the camera attached to a unit, and will show everything from
// his viewpoint
	void AttachToUnit(WarUnit *body);
	void DetachFromUnit();

// makes the camera lock it's current relative position to the unit
	void LockToUnit(WarUnit *body);
	void UnlockFromUnit();

// makes the camera look on a unit and follow it
// only rotates the camera, don't change the tilt angle
	void LookAtUnit(WarUnit *body);
	void StopLookAtUnit();

// changes Field of vision
	void SetFov(float horizontal_fov,float vertical_fov);
	float HorizontalFov();
	float VerticalFov();

// changes the far z clip distance
	float FarClip();
	void SetFarClip(float far_clip);

// rotates the camera to look at the given point
	void LookAt(WarVertex point);

	bool IsBoxVisible(WarBox *box,BRect rect);
	bool IsBoxVisible(WarBox *box);
	bool IsBoxVisible(WarBox *box,bool *entire);

private:

	WarVertex m_movement_speed;
	WarVertex m_rotation_speed;
	float m_vertical_move_speed;
	float m_horizontal_move_speed;

	float m_time_left; // the time left before the camera must have reached it's goal (camera_goal)
	WarVertex m_camera_goal; // the camera_goal, ignore it if time_left is 0

	float m_hfov; // horizontal field of vision
	float m_vfov; // vertical field of vision
	float m_far_z_clip; // the distance to the far z clip plane

	WarVertex m_normal[6]; // normals for the four frustum planes, far z clip and near z clip plane, not rotated
	WarVertex m_rotated_normal[6]; // normals for the four frustum planes, far z clip and near z clip plane, rotated
	float m_plane_d[6];			// the d constant for each plane
	
	// the attributes the last time cycle was called
	float m_safe_hfov;
	float m_safe_vfov;
	WarVertex m_safe_position;
	WarVertex m_safe_direction;
	float m_safe_far_z_clip;

	// LOCKED TO UNIT -------------------------------

	WarUnit *m_locked_unit; // is NULL if not locked
	// the id of the unit the camera is locked to, we need it since we don't know
	// when the Body is deleted, we have to check if it still exists every cycle
	uint16 m_locked_unit_id;
	WarVertex m_relative_position; // stores the relative position between the unit and the camera

	// ATTACHED TO UNIT -------------------------------

	WarUnit *m_attached_unit; // is NULL if not attached
	// the id of the unit the camera is attached to, we need it since we don't know
	// when the Body is deleted, we have to check if it still exists every cycle
	uint16 m_attached_unit_id;
	
	// LOOK AT UNIT -------------------------------
	
	WarUnit *m_look_at_unit; // is NULL if not looked at
	// the id of the unit the camera is attached to, we need it since we don't know
	// when the Body is deleted, we have to check if it still exists every cycle
	uint16 m_look_at_unit_id;
};

// inline definitions

inline void WarCamera::Stop()
{
	m_vertical_move_speed = 0.0;
	m_horizontal_move_speed = 0.0;
	m_movement_speed.SetTo(0.0,0.0,0.0);
	m_rotation_speed.SetTo(0.0,0.0,0.0);
}

inline void WarCamera::AddScrollSpeed(float horiz,float vert)
{
	m_horizontal_move_speed += horiz;
	m_vertical_move_speed += vert;
}

inline void WarCamera::AddMovementSpeed(WarVertex speed)
{
	m_movement_speed += speed;
}

inline void WarCamera::AddRotationSpeed(WarVertex speed)
{
	m_rotation_speed += speed;
//	printf("rotation_speed: (%f %f %f)\n",rotation_speed.x,rotation_speed.y,rotation_speed.z);
}

inline void WarCamera::SetFov(float horizontal_fov,float vertical_fov)
{
	if (horizontal_fov > 1.141) m_hfov=1.141;
	else m_hfov = horizontal_fov;
	if (vertical_fov > 1.141) m_vfov=1.141;
	else m_vfov = vertical_fov;
}

inline bool WarCamera::IsBoxVisible(WarBox *box,BRect rect)
{
	rect.left *= 2;
	rect.top *= 2;
	rect.right *= 2;
	rect.bottom *= 2;
	rect.OffsetBy(-1.0,-1.0);

	float h = -HorizontalFov();
	float v = VerticalFov();

	WarVertex normal[6];
	float angle;

// set up the frustum plane normals
	// left plane
	angle = atan((tan(h/2.0)*rect.left)) + 3.1415927/2.0;
	normal[0].SetTo(cos(angle),sin(angle),0.0);
	// right plane
	angle = atan((tan(h/2.0)*rect.right)) - 3.1415927/2.0;
	normal[1].SetTo(cos(angle),sin(angle),0.0);
	// top plane
	angle = atan((tan(v/2.0)*rect.top)) + 3.1415927/2.0;
	normal[2].SetTo(cos(angle),0.0,sin(angle));
	// bottom plane
	angle = atan((tan(v/2.0)*rect.bottom)) - 3.1415927/2.0;
	normal[3].SetTo(cos(angle),0.0,sin(angle));
	// the far_clip & near_clip planes
	normal[4].SetTo(m_far_z_clip,0.0,0.0);
	normal[5].SetTo(-0.1,0.0,0.0);


	// checking against the first four planes (left,right,top,bottom)
	for (int32 i=0;i<4;i++)
	{
		normal[i].RotateBy(Direction());
		float d = - normal[i].x * Position().x - normal[i].y * Position().y - normal[i].z * Position().z;

		int32 j=0;
		while (j < 8 && normal[i].x * box->Point(j)->x + normal[i].y * box->Point(j)->y + normal[i].z * box->Point(j)->z + d > 0) j++;
		if (j == 8) return false;
	}

	// checking against the far & near z clip plane
	for (int32 i=4;i<6;i++)
	{
		normal[i].RotateBy(Direction());
		float d = - normal[i].x * (normal[i].x + Position().x) - normal[i].y * (normal[i].y + Position().y) - normal[i].z * (normal[i].z + Position().z);
		int32 j=0;
		while (j < 8 && normal[i].x * box->Point(j)->x + normal[i].y * box->Point(j)->y + normal[i].z * box->Point(j)->z + d > 0) j++;
		if (j == 8) return false;
	}

	return true;
}

inline bool WarCamera::IsBoxVisible(WarBox *box, bool *entire)
{
	int32 c;

	*entire = true;

	// checking against the first four planes (left,right,top,bottom)

	for (int32 i=0;i<6;i++)
	{
		c = 0;
		for (int32 j=0;j<8;j++)
			if (m_rotated_normal[i].x * box->Point(j)->x + m_rotated_normal[i].y * box->Point(j)->y + m_rotated_normal[i].z * box->Point(j)->z + m_plane_d[i] > 0) c++;
		if (c > 0) *entire = false;
		if (c == 8) return false;
	}

	return true;
}

inline bool WarCamera::IsBoxVisible(WarBox *box)
{
	int32 j;

	// checking against the first four planes (left,right,top,bottom)
	for (int32 i=0;i<6;i++)
	{
		j=0;
		while (j < 8 && m_rotated_normal[i].x * box->Point(j)->x + m_rotated_normal[i].y * box->Point(j)->y + m_rotated_normal[i].z * box->Point(j)->z + m_plane_d[i] > 0) j++;
		if (j == 8) return false;
	}

	return true;
}

inline float WarCamera::FarClip()
{
	return m_safe_far_z_clip;
}

inline void WarCamera::SetFarClip(float far_clip)
{
	m_far_z_clip = far_clip;
}

inline float WarCamera::HorizontalFov()
{
	return m_safe_hfov;
}

inline float WarCamera::VerticalFov()
{
	return m_safe_vfov;
}

#endif // #ifndef _WAR_CAMERA_