#include "WarCamera.h"
#include "WarGLView.h"
#include "WarMap.h"
#include "WarBox.h"
#include "defs.h"

#include <TranslationUtils.h>
#include <List.h>
#include <Rect.h>
#include <stdlib.h>
#include <math.h>
#include <gl.h>
#include <glu.h>

WarCamera::WarCamera():WarMatrix()
{
	m_attached_unit=NULL;
	m_look_at_unit=NULL;
	SetFov(3.1415927/2.0,3.1415927/2.0);
	SetFarClip(200.0);
	m_time_left = 0.0;
	m_vertical_move_speed = 0.0;
	m_horizontal_move_speed = 0.0;
	
	m_safe_hfov = 0.0;
	m_safe_vfov = 0.0;

	m_safe_far_z_clip = 0.0;
}

WarCamera::WarCamera(WarVertex camera_position):WarMatrix(camera_position)
{
	m_attached_unit=NULL;
	m_look_at_unit=NULL;
	SetFov(3.1415927/2.0,3.1415927/2.0);
	SetFarClip(200.0);
	m_time_left = 0.0;
	m_vertical_move_speed = 0.0;
	m_horizontal_move_speed = 0.0;

	m_safe_hfov = 0.0;
	m_safe_vfov = 0.0;

	m_safe_far_z_clip = 0.0;
}

WarCamera::~WarCamera()
{
}

void WarCamera::Scroll(float horiz,float vertical)
{
	if (horiz)
	{
		m_position.x+=sin(m_direction.z)*horiz;
		m_position.y-=cos(m_direction.z)*horiz;
	}

	if (vertical)
	{
		m_position.x+=cos(m_direction.z)*vertical;
		m_position.y+=sin(m_direction.z)*vertical;
	}
}

void WarCamera::LockToUnit(WarUnit *unit)
{
	m_locked_unit = unit;

	if (m_locked_unit != NULL)
	{
		m_locked_unit_id = m_locked_unit->Id();
		m_relative_position = m_position - m_locked_unit->Position();
	}
}

void WarCamera::UnlockFromUnit()
{
	m_locked_unit = NULL;
}

void WarCamera::AttachToUnit(WarUnit *unit)
{
	m_attached_unit = unit;

	if (m_attached_unit != NULL)
	{
		m_attached_unit_id = m_attached_unit->Id();
		MoveTo(m_attached_unit->Position());
		RotateTo(m_attached_unit->Direction());
	}
}

void WarCamera::DetachFromUnit()
{
	m_attached_unit = NULL;
}

void WarCamera::LookAtUnit(WarUnit *unit)
{
	m_look_at_unit = unit;
	if (m_look_at_unit != NULL)
	{
		m_look_at_unit_id = m_look_at_unit->Id();
		LookAt(m_look_at_unit->Position());
	}
}

void WarCamera::StopLookAtUnit()
{
	m_look_at_unit = NULL;
}

void WarCamera::Cycle(float frame_time,WarEntityList *war_entity_list,WarMap *war_map)
{
	if (!IsLocked())
	{
		if (m_time_left)
		{
			if (m_time_left<frame_time)
			{
				m_position = m_camera_goal;
				m_time_left=0;
			}
			else
			{
				WarVertex tmp=(m_camera_goal - m_position) * (frame_time/m_time_left);
				m_position += tmp;
				m_time_left-= frame_time;
			}
		}

		if (m_horizontal_move_speed || m_vertical_move_speed)
			Scroll(m_horizontal_move_speed * frame_time,m_vertical_move_speed * frame_time);

		if (m_movement_speed != WarVertex(0.0,0.0,0.0))
			MoveBy(m_movement_speed * frame_time);

		if (m_rotation_speed != WarVertex(0.0,0.0,0.0))
			RotateBy(m_rotation_speed * frame_time);

		if (m_attached_unit != NULL && war_entity_list->GetBodyByID(m_attached_unit_id) == m_attached_unit)
		{
			MoveTo(m_attached_unit->Position());
			RotateTo(m_attached_unit->Direction());
		}

		if (m_locked_unit != NULL && war_entity_list->GetBodyByID(m_locked_unit_id) == m_locked_unit)
			MoveTo(m_locked_unit->Position() + m_relative_position);

		if (m_look_at_unit != NULL && war_entity_list->GetBodyByID(m_look_at_unit_id) == m_look_at_unit)
			LookAt(m_look_at_unit->Position());
	}
}

void WarCamera::LookAt(WarVertex point)
{
	float dist_x = point.x - m_position.x;
	float dist_y = point.y - m_position.y;
	float dist_z = point.z - m_position.z;

	RotateZTo(atan2(dist_y,dist_x));
	RotateYTo(atan2(-dist_z,sqrt(dist_y*dist_y + dist_x*dist_x)));
	// dist_z have to be negative because the y-angles are positive downwards
}

void WarCamera::Swap()
{
	bool cont = false;

	if (m_safe_vfov != m_vfov || m_safe_hfov != m_hfov || m_far_z_clip != m_safe_far_z_clip)
	{
		m_safe_far_z_clip = m_far_z_clip;
		m_safe_hfov = m_hfov;
		m_safe_vfov = m_vfov;

// recalculate the frustum plane normals
		float h = -m_hfov;
		float v = m_vfov;
		float angle;
		// left plane
		angle = atan(-tan(h/2.0)) + 3.1415927/2.0;
		m_normal[0].SetTo(cos(angle),sin(angle),0.0);
		// right plane
		angle = atan(tan(h/2.0)) - 3.1415927/2.0;
		m_normal[1].SetTo(cos(angle),sin(angle),0.0);
		// top plane
		angle = atan(tan(v/2.0)) + 3.1415927/2.0;
		m_normal[2].SetTo(cos(angle),0.0,sin(angle));
		// bottom plane
		angle = atan(-tan(v/2.0)) - 3.1415927/2.0;
		m_normal[3].SetTo(cos(angle),0.0,sin(angle));
		// far z plane
		m_normal[4].SetTo(FarClip(),0.0,0.0);
		// near z plane
		m_normal[5].SetTo(-1.0,0.0,0.0);
		
		cont = true;
	}

	if (m_safe_direction != m_direction || cont)
	{
		m_safe_direction = m_direction;
		for (int32 i=0;i<6;i++)
		{
			m_rotated_normal[i] = m_normal[i];
			m_rotated_normal[i].RotateBy(Direction());
		}
		cont = true;
	}

	if (m_safe_position != m_position || cont)
	{
		m_safe_position = m_position;
		for (int32 i=0;i<4;i++)
			m_plane_d[i] = - m_rotated_normal[i].x * XPosition() - m_rotated_normal[i].y * YPosition() - m_rotated_normal[i].z * ZPosition();

		m_plane_d[4] = - m_rotated_normal[4].x * (m_rotated_normal[4].x + Position().x) - m_rotated_normal[4].y * (m_rotated_normal[4].y + Position().y) - m_rotated_normal[4].z * (m_rotated_normal[4].z + Position().z);
		m_plane_d[5] = - m_rotated_normal[5].x * (-m_rotated_normal[5].x + Position().x) - m_rotated_normal[5].y * (-m_rotated_normal[5].y + Position().y) - m_rotated_normal[5].z * (-m_rotated_normal[5].z + Position().z);
	}
}

_EXPORT WarVertex WarCamera::Position()
{
	return m_safe_position;
}

_EXPORT float WarCamera::XPosition()
{
	return m_safe_position.x;
}

_EXPORT float WarCamera::YPosition()
{
	return m_safe_position.y;
}

_EXPORT float WarCamera::ZPosition()
{
	return m_safe_position.z;
}

_EXPORT WarVertex WarCamera::Direction()
{
	return m_safe_direction;
}

_EXPORT float WarCamera::XDirection()
{
	return m_safe_direction.x;
}

_EXPORT float WarCamera::YDirection()
{
	return m_safe_direction.y;
}

_EXPORT float WarCamera::ZDirection()
{
	return m_safe_direction.z;
}
