#include "WarGLView.h"
#include "WarWindow.h"
#include "WarStatView.h"
#include "WarDisplay.h"
#include "WarStack.h"
#include "WarStringInput.h"
#include "WarMenu.h"
#include "WarView.h"
#include "WarSkydome.h"
#include "WarWeather.h"
#include "WarCli.h"
#include "WarTextOutput.h"

#include <TranslationUtils.h>
#include <Autolock.h>
#include <opengl/GL/gl.h>
#include <opengl/GL/glu.h>
#include <app/Application.h>
#include <interface/InterfaceDefs.h>
#include <Directory.h>
#include <Path.h>

#include <stdio.h>

long WarGLView::main_task(void *data)
{
	WarGLView *view=(WarGLView*)data;

	bigtime_t time=system_time();
	bigtime_t systime;
	float frame_time=0.01;

	// main loop
	while (view->m_running)
	{
		view->Cycle(frame_time);
		view->Render();
		systime = system_time();

		// limit the frame rate to 50 fps
		if (systime - time < 20000 )
		{
			snooze(20000 - systime + time);
			systime = system_time();
		}

		frame_time = (float)(systime - time) / 1000000.0;
		time = systime;
	}
	return 0;
}

WarGLView::WarGLView( BRect frame, char *name, int32 mode, int32 flags )
	: BGLView( frame, name, mode, flags, BGL_RGB | BGL_DOUBLE | BGL_DEPTH )
{
// temporary, will be set during connecting to a server
	m_team=0;

// we want keyboard events even when we're not in focus
//	SetEventMask(B_KEYBOARD_EVENTS);
	SetEventMask(B_POINTER_EVENTS,B_NO_POINTER_HISTORY);


	m_running = true;

// create the word list
	m_word_list=new WarList<char*>(100,20);

// construct the stacks
	m_integer_stack = new WarStack<int32>(stack_size);
	m_float_stack = new WarStack<double>(stack_size);
	m_string_stack = new WarStack<BString*>(stack_size);

// construct the focus stack
	m_focus_stack = new WarStack<WarView*>(10);

// construct the view-remove stack
	m_remove_stack = new WarStack<WarView*>(1);

// copy the word list into the WarList
	for (uint16 i=0;i<num_words;i++)
	{
		char *word = new char[strlen(words[i])+1];
		strcpy(word,words[i]);
		m_word_list->AddItem(word);
	}

	app_info info;
	be_app->GetAppInfo(&info);
	BEntry app_entry(&info.ref);
	BPath app_path(&app_entry);
	app_path.GetParent(&app_path);

	m_app_dir = new BDirectory(app_path.Path());
	m_script_dir = new BDirectory(m_app_dir,"scripts/");
	m_environment_dir = new BDirectory(m_app_dir,"environments/");
	m_map_dir = new BDirectory(m_app_dir,"maps/");
	m_unit_dir = new BDirectory(m_app_dir,"units/");

// allocating the camera list
	m_camera = new WarCamera[16];
	m_current_camera = m_camera;
	
	for (int32 i=0;i<16;i++)
		m_camera[i].MoveTo(WarVertex(30,30,100));

// initializing the keyboard_state array
	for (int32 i=0;i<130;i++)
		m_keyboard_state[i]=false;
}

WarGLView::~WarGLView( void )
{
	delete m_war_entity_list;
	delete m_war_map;
	delete m_war_skydome;
	delete m_war_weather;
	delete m_war_display;
	delete m_war_cli;
	delete m_war_statview;
	delete m_war_text_output;
	delete m_war_selection_view;
	delete m_bigfont;
	delete m_smallfont;

// delete all strings in the wordlist
	uint16 cookie = 0;
	char *word;
	while (m_word_list->GetNextItem(&word,&cookie))
		delete []word;
	
	delete m_word_list;

	delete m_integer_stack;
	delete m_float_stack;
	delete m_string_stack;

	delete m_remove_stack;
	delete m_focus_stack;

	delete m_unit_dir;
	delete m_map_dir;
	delete m_environment_dir;
	delete m_script_dir;
	delete m_app_dir;
}

void WarGLView::AddChild(WarView *view)
{
	view->m_gl_view=this;
	m_childs.AddItem(view);
}

void WarGLView::RemoveChild(WarView *view)
{
	printf("WarGLView::RemoveChild() trying to lock child lock\n");
	BAutolock lock(m_child_lock);
	if (lock.IsLocked())
	{
		// remove the view from teh focus stack
		printf("WarGLView::RemoveChild() LeaveFocus()\n");
		LeaveFocus(view);
		// maybe it wasn't at the top position
		printf("WarGLView::RemoveChild() focus_view->remove(view)\n");
		if (m_focus_stack->CountElements()) m_focus_stack->Remove(view);

		printf("WarGLView::RemoveChild() childs->RemoveItem(view)\n");
		m_childs.RemoveItem(view);

		printf("WarGLView::RemoveChild() unlock\n");
	}
	else printf("WarGLView::RemoveChild() for some reason the lock couldn't be locked\n");
}

void WarGLView::DetachedFromWindow( void )
{
	BGLView::DetachedFromWindow();

// Don't ask me why I have to do this,
// I just copied it from the GLTeapot sample
	long locks=0;
	while (Window()->IsLocked())
	{
		locks++;
		Window()->Unlock();
	}

	int32 ret;
	m_running = false;
	wait_for_thread(m_main_thread,&ret);

	while(locks--) Window()->Lock();
}

void WarGLView::LeaveFocus(WarView *view)
{
	if (m_focus_stack->CountElements() == 0) return;

	WarView *focus_view;

	focus_view = m_focus_stack->Top();
	if (view == focus_view)
	{
		// the given view was the focus view, remove it from the focus stack
		focus_view = m_focus_stack->Pop();
		// and tell the view it lost it's focus
		focus_view->LostFocus();

		// tell the old view it got the focus back
		if (!m_focus_stack->IsEmpty())
		{
			focus_view = m_focus_stack->Top();
			focus_view->GotFocus();
		}
	}
}

void WarGLView::GiveFocus(WarView *view)
{
	WarView *focus_view = NULL;
	
	// first, tell the current focus view it is loosing it's focus
	if (!m_focus_stack->IsEmpty())
	{
		focus_view = m_focus_stack->Top();
		focus_view->LostFocus();
	}
	// push the given view to the focus stack
	m_focus_stack->Push(view);
	// and tell it that it got the focus
	view->GotFocus();
}

void WarGLView::SetMousePosition(WarPoint position)
{
	BPoint point;
	point.y = Frame().Height() - position.y;
	point.x = position.x;

	ConvertToScreen(&point);
//	printf("GLView screen: set_mouse_position: (%d,%d)\n",(int32)position.x,(int32)position.y);
	set_mouse_position((int32)point.x,(int32)point.y);
}

void WarGLView::AttachedToWindow( void )
{
	// Make keyboard/mouse input go to this view when its window has
	// focus (that'll stay in effect until another view calls MakeFocus()
	// or the user puts the focus on a control... which won't happen in
	// this app because there are no other views and no controls).
	MakeFocus();

	// Initialize OpenGL; gInit() sets things up, gReshape() makes sure
	// OpenGL knows about the view's size/shape.
	LockGL();
	BGLView::AttachedToWindow();

	// Set up the depth buffer
	glClearDepth( 1.0 );
	glDepthFunc( GL_LESS );
	glEnable( GL_DEPTH_TEST );

	BRect r = Bounds();

	m_bigfont = new WarFont(m_app_dir,"bigfont");
	m_smallfont = new WarFont(m_app_dir,"smallfont");

	m_war_map=new WarMap(m_map_dir,"default",m_environment_dir);
	m_war_map->GenerateMap();

	m_war_selection_view = new WarSelectionView(r.IntegerWidth()>>1,r.IntegerHeight()>>1,r.IntegerWidth()>>1,r.IntegerHeight()>>1,m_smallfont);
	m_war_entity_list = new WarEntityList(m_war_map, m_war_selection_view, m_unit_dir);

	m_war_weather = new WarWeather("sky.jpg",WarColor(40,132,255));
//	m_war_weather = new WarWeather("nightsky.jpg",WarColor(0,0,0));

	m_war_skydome = new WarSkydome(m_war_weather,1.0);

	m_planet = new WarPlanet(WarEntity(0,100.0,WarVertex(0,0,-2300.0),WarVertex(0,0,.001)),2000,m_war_map, m_war_entity_list, m_war_weather);

	m_war_display = new WarDisplay(0,0,r.IntegerWidth(),r.IntegerHeight(), m_planet, m_war_skydome, m_current_camera, m_team);
	AddChild(m_war_display);
	AddChild(m_war_selection_view);

	m_war_statview = new WarStatView(0.0,32.0,128.0,64.0,m_smallfont);
	AddChild(m_war_statview);

	m_war_text_output = new WarTextOutput(0,r.IntegerHeight()>>1,r.IntegerWidth()>>1,r.IntegerHeight()>>1,m_smallfont);
	AddChild(m_war_text_output);

	m_war_cli=new WarCli(0,0,r.IntegerWidth(),32,m_smallfont,m_word_list);
	AddChild(m_war_cli);
	m_war_cli->SetPrompt("0 > ");
	GiveFocus(m_war_cli);

	m_war_string_input=new WarStringInput(0,m_smallfont->Height() + 1,r.IntegerWidth(),32,m_smallfont,m_word_list);
	AddChild(m_war_string_input);
	m_war_string_input->Hide(true);

	m_war_mouse_view = new WarMouseView(0,0,r.IntegerWidth(),r.IntegerHeight());
	AddChild(m_war_mouse_view);
	RemoveChild(m_war_mouse_view);

	char string[16];
	strcpy(string,"\" startup\" include");
	Parse(string);

	// Clear to the fog color
	glClearColor(m_war_weather->FogColor()[0],m_war_weather->FogColor()[1],m_war_weather->FogColor()[2],m_war_weather->FogColor()[3]);

	// makes sure the backfaces are culled
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	UnlockGL();

//----------

	uint16 class_id = m_war_entity_list->LoadClass("tank1");
	uint16 entity_id;

	entity_id = m_war_entity_list->AddEntity(140.0,110.0,10.0,class_id,0,0);
	m_war_entity_list->TakeOff(entity_id);

	entity_id = m_war_entity_list->AddEntity(110.0,110.0,10.0,class_id,0,0);
	m_war_entity_list->TakeOff(entity_id);

	entity_id = m_war_entity_list->AddEntity(130.0,120.0,10.0,class_id,0,0);
	m_war_entity_list->TakeOff(entity_id);

//	printf("entity_id: %d\n",entity_id);

	for (uint16 i=0;i<16;i++)
	{
		m_camera[i].MoveTo(WarVertex(10.0,10.0,20.0));
		m_camera[i].RotateYTo(3.1415927/2);
	}

	m_current_camera->MoveTo(WarVertex(100.0,100.0,20.0));
	m_current_camera->RotateYTo(3.1415927/2);

//----------

	m_main_thread = spawn_thread(main_task,"main_thread",B_NORMAL_PRIORITY,(void*)this);
	resume_thread(m_main_thread);
}

void WarGLView::MouseMoved(BPoint point,uint32 transit,const BMessage *message)
{
	BAutolock lock(m_child_lock);
	if (!lock.IsLocked()) return;

	point.x = point.x;
	point.y = Frame().Height() - point.y;

	WarView *view;
	uint16 cookie = 0;
	while (m_childs.GetNextItem(&view,&cookie))
	{
		WarPoint mouse_pos;
		mouse_pos = point;

		mouse_pos.x -= view->Left();
		mouse_pos.y -= view->Top();
	
		view->MouseMoved(mouse_pos);
	}

	WarPoint mouse_pos;
	mouse_pos = point;
	mouse_pos.x -= m_war_mouse_view->Left();
	mouse_pos.y -= m_war_mouse_view->Top();
	m_war_mouse_view->MouseMoved(mouse_pos);
}

void WarGLView::MouseDown(BPoint point)
{
	if (Bounds().Contains(point) && Window()->IsActive())
	{
		BAutolock lock(m_child_lock);
		if (!lock.IsLocked()) return;

		BMessage *message = Window()->CurrentMessage();
		int32 buttons;
		message->FindInt32("buttons",&buttons);

		point.y = Frame().Height() - point.y;

		WarView *view;
		uint16 cookie = 0;
		uint8 pressed_buttons = buttons ^ m_mouse_buttons;
		while (m_childs.GetNextItem(&view,&cookie))
		{
			WarPoint mouse_pos;
			mouse_pos = point;

			mouse_pos.x -= view->Left();
			mouse_pos.y -= view->Top();

			view->MouseDown(mouse_pos,pressed_buttons);
		}
		m_mouse_buttons = buttons;
	}
}

void WarGLView::MouseUp(BPoint point)
{
	BMessage *message = Window()->CurrentMessage();
	int32 buttons;
	message->FindInt32("buttons",&buttons);

	if (buttons != m_mouse_buttons)
	{
		BAutolock lock(m_child_lock);
		if (!lock.IsLocked()) return;

		point.y = Frame().Height() - point.y;

		WarView *view;
		uint16 cookie = 0;
		uint8 released_buttons = buttons ^ m_mouse_buttons;
		while (m_childs.GetNextItem(&view,&cookie))
		{
			WarPoint mouse_pos;
			mouse_pos = point;

			mouse_pos.x -= view->Left();
			mouse_pos.y -= view->Top();

			view->MouseUp(mouse_pos,released_buttons);
		}

		m_mouse_buttons = buttons;
	}
}

void WarGLView::FrameResized( float width, float height )
{
	// When the view is resized (it follows the window, since we created
	// it with B_FOLLOW_ALL_SIDES), we need to tell OpenGL that it's
	// rendering into a new size/shape.
	LockGL();
	BGLView::FrameResized( width, height );
	UnlockGL();

	// The OpenGL view has been resized, so we have to reconfigure our
	// OpenGL context to reflect that.
	m_war_display->Resize(width,height);

//	m_war_display2->Resize(width/2,height);
//	m_war_display2->Move((int)width/2,0);

	m_war_cli->Resize(width,32);

	m_war_text_output->Move(0,height/2);
	m_war_text_output->Resize(width/2,height/2);

	m_war_selection_view->Move(width/2,height/2);
	m_war_selection_view->Resize(width/2,height/2);
	
	m_war_mouse_view->Resize(width,height);

//	Render();
}

void WarGLView::KeyDown( const char *bytes, int32 size )
{
	BAutolock lock(m_child_lock);
	if (!lock.IsLocked()) return;

//	if (!Window()->IsActive()) return;
	BMessage *message = Window()->CurrentMessage();

//	message->PrintToStream();

	int32 key;
	message->FindInt32("key",&key);

	if (m_focus_stack->IsEmpty())
	{
		if (m_keyboard_state[key]) return;
		m_keyboard_state[key]=true;

		if (key == 17)
			Parse("+paragraph",true);
		else if (key == 1)
			Parse("+escape",true);
		// now, we should execute the function assigned to the pressed key
		else if (size == 1)
		{
			char tmp[5];
			if (bytes[0] == B_UP_ARROW)
				Parse("+up_arrow",true);
			else if (bytes[0] == B_DOWN_ARROW)
				Parse("+down_arrow",true);
			else if (bytes[0] == B_LEFT_ARROW)
				Parse("+left_arrow",true);
			else if (bytes[0] == B_RIGHT_ARROW)
				Parse("+right_arrow",true);
			else if ((bytes[0] >= '0' && bytes[0] <= '9') || bytes[0] >= 'a' && bytes[0] <= 'z')
			{
				tmp[0]='+';
				tmp[1]=bytes[0];
				tmp[2]='\0';

				Parse(tmp,true);
			}
			else if (bytes[0] == B_FUNCTION_KEY && key >= B_F1_KEY && key <= B_F12_KEY)
			{
				sprintf(tmp,"+f%ld",key - B_F1_KEY + 1);
				Parse(tmp,true);
			}
		}
	}
	else
	{
		WarView *focus_view;
		focus_view = m_focus_stack->Top();

		focus_view->KeyDown(*bytes);
	}

	BGLView::KeyDown( bytes, size );
}

void WarGLView::KeyUp( const char *bytes, int32 size )
{
	BAutolock lock(m_child_lock);
	if (!lock.IsLocked()) return;

	BMessage *message = Window()->CurrentMessage();

//	message->PrintToStream();

	int32 key;
	message->FindInt32("key",&key);

//	printf("key: %ld\n",key);

	if (m_keyboard_state[key])
	{
		m_keyboard_state[key]=false;

		if (key == 17)
			Parse("-paragraph",true);
		else if (key == 1)
			Parse("-escape",true);
		// now, we should execute the function assigned to the released key
		else if (size == 1)
		{
			char tmp[5];
			if (bytes[0] == B_UP_ARROW)
				Parse("-up_arrow",true);
			else if (bytes[0] == B_DOWN_ARROW)
				Parse("-down_arrow",true);
			else if (bytes[0] == B_LEFT_ARROW)
				Parse("-left_arrow",true);
			else if (bytes[0] == B_RIGHT_ARROW)
				Parse("-right_arrow",true);
			else if ((bytes[0] >= '0' && bytes[0] <= '9') || bytes[0] >= 'a' && bytes[0] <= 'z')
			{
				tmp[0]='-';
				tmp[1]=bytes[0];
				tmp[2]='\0';

				Parse(tmp,true);
			}
			else if (bytes[0] == B_FUNCTION_KEY && key >= B_F1_KEY && key <= B_F12_KEY)
			{
				sprintf(tmp,"-f%ld",key - B_F1_KEY + 1);
				Parse(tmp,true);
			}
		}
	}

	if (!m_focus_stack->IsEmpty())
	{
		WarView *focus_view;
		focus_view = m_focus_stack->Top();
		focus_view->KeyUp(bytes[0]);
	}

	BGLView::KeyDown( bytes, size );
}

void WarGLView::ErrorCallback( GLenum err )
{
	// The ErrorCallback() method is called whenever OpenGL has a problem;
	// you can override it to ignore specific known error messages and
	// display otheres, abort the program, etc... we'll just print them
	// out.  You'll have to run the application from a Terminal to see
	// these.
	fprintf( stderr, "Unexpected error occured (%ld):\n", err );
	fprintf( stderr, "\t%s\n", gluErrorString( err ) );
}

void WarGLView::Cycle(float frame_time)
{
// cycle the entities
	WarUnit *entity;
	uint16 i,cookie=0;
	while (m_war_entity_list->GetNextEntity(&entity,NULL,&i,&cookie))
	{
		entity->Cycle(NULL,frame_time);
//		if (entity->Team() == m_team) entity->Vision(m_war_map,MAPSTATE_VISIBLE | MAPSTATE_DISCOVERED);
	}

// Cycle the cameras, and make sure they're not outside the map
	for (uint16 i=0;i<16;i++)
	{
//		bool changed = false;
		m_camera[i].Cycle(frame_time,m_war_entity_list,m_war_map);
		m_camera[i].Swap();
/*
		if (m_camera[i].XPosition() > m_war_map->Width() - 1.0)
		{
			m_camera[i].MoveXTo(m_war_map->Width() - 1.0);
			changed = true;
		}
		else if (m_camera[i].XPosition() < 0.0)
		{
			m_camera[i].MoveXTo(0.0);
			changed = true;
		}

		if (m_camera[i].YPosition() > m_war_map->Height() - 1.0)
		{
			m_camera[i].MoveYTo(m_war_map->Height() - 1.0);
			changed = true;
		}
		else if (m_camera[i].YPosition() < 0.0)
		{
			m_camera[i].MoveYTo(0.0);
			changed = true;
		}

		float map_height = m_war_map->GetHeight(m_camera[i].XPosition(),m_camera[i].YPosition()) + 1.0;

		if (m_camera[i].ZPosition() < map_height)
		{
			m_camera[i].MoveZTo(map_height);
			changed = true;
		}
		else if (m_camera[i].ZPosition() > 255.0)
		{
			m_camera[i].MoveZTo(255.0);
			changed = true;
		}

		if (changed) m_camera[i].Swap();
*/
	}

	m_planet->Cycle(frame_time);

// update the status view info
	time_accum += frame_time;
	frames++;
	if (time_accum >= .3)
	{
		m_war_statview->SetFps(((float)frames)/time_accum);
		frames = 0;
		time_accum = 0.0;
	}
	m_war_statview->SetFans(m_war_display->FanCount());
	m_war_statview->SetPolys(m_war_display->PolygonCount());

// remove every view that should be removed
	while (m_remove_stack->CountElements())
	{
		printf("WarGLView::Cycle() pop from remove_stack\n");
		WarView *view = m_remove_stack->Pop();

		// now the view can't be accessed by anybody else, it's not stored in the focus stack
		// and is not in the child list.
		printf("WarGLView::Cycle() RemoveChild()\n");
		RemoveChild(view);
		printf("WarGLView::Cycle() delete view\n");
		delete view;
	}
}

void WarGLView::Render(void)
{
	// Render our OpenGL scene and swap the buffer so we can actually see
	// it.

	LockGL();
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

// render all the normal views
	WarView *view;
	uint16 cookie = 0;
	while (m_childs.GetNextItem(&view,&cookie))
		view->Invalidate();

// render the mouse view, this should always be rendered last, that's why
// it's not in the regular view list
	m_war_mouse_view->Invalidate();

	SwapBuffers();
	UnlockGL();
}

void WarGLView::Parse(char *line,bool quiet,int32 call_stack)
{
	if (*line == 0) return;
/*
	// ********************** COMPILE ********************

	// if the user is trying to compile
	if (!strncmp(line,words[WORD_COMPILE],strlen(words[WORD_COMPILE])) && line[strlen(words[WORD_COMPILE])]==' ' && strlen(line)-strlen(words[WORD_COMPILE])>1)
	{
	}

	// ********************** FORGET ********************
	
	// if the user is trying to forget
	if (!strncmp(line,words[WORD_FORGET],strlen(words[WORD_FORGET])) && line[strlen(words[WORD_FORGET])]==' ')
	{
		char *defined_word;
		char *unwanted_word=line+strlen(words[WORD_FORGET])+1;

		// prepare the parameter
		uint16 i=0;
		while (unwanted_word[i]!=' ' && unwanted_word[i]!=0)
			i++;
		unwanted_word[i]=0;

		//  look for the word to forget
		i=0;
		bool is_word_defined;
		while (is_word_defined = m_word_list->GetNextItem(&defined_word,&i))
			if (!strcmp(unwanted_word,defined_word)) break;
		i--;

		if (!is_word_defined)
		{ // the wor does not exist in the word list, it is not defined. Tell the user and return
			char s[256];
			sprintf(s,"the word \"%s\" is not defined",unwanted_word);
			m_war_text_output->PutLine(s);
			return;
		}

		if (i<num_words)
		{ // the word is not user-defined and cannot be forgotten. Tell the user and return.
			char s[256];
			sprintf(s,"the word \"%s\" is not user defined and cannot be forgotten",unwanted_word);
			m_war_text_output->PutLine(s);
			return;
		}
		
		defined_word = m_word_list->RemoveItemAt(i);
		delete []defined_word;
		return;
	}
	
*/
	// ************ STRING *********************

	try
	{
		char *word;
		char *command_line=NULL;
		word = GetWord(line,&command_line);
		do
		{
			if (call_stack < 100) // 100 recursive calls, that's all
			{
				int32 len = strlen(command_line);
				int32 jump = execute(word,command_line,quiet,call_stack);
				if (jump <= len) command_line += jump;
				else command_line += len;
			}
			else
			{
				m_war_text_output->PutLine("call stack overflow, terminating.\nYou can only have 100 recursive function calls.");
			
				char prompt[10];
				sprintf(prompt,"%ld > ",m_integer_stack->CountElements());
				m_war_cli->SetPrompt(prompt);
				break;
			}
			word = GetWord(line,&command_line);
		}while (word);

	}
	catch (WarException *exception)
	{
		if (exception->HasMessage())
		{
			printf("caught exception: \"%s\"\n",exception->Message());
			m_war_text_output->PutLine(exception->Message());
		}
		delete exception;
	}
	catch (...)
	{
		m_war_text_output->PutLine("Unknown exception");
	}


	char prompt[10];
	sprintf(prompt,"%ld > ",m_integer_stack->CountElements());
	m_war_cli->SetPrompt(prompt);
}

WarMenu *WarGLView::ParseMenu(WarPoint position,char *command_line)
{
	int32 num_params = m_integer_stack->Pop();
	if (num_params)
	{
		WarMenu *menu = new WarMenu(position.x,position.y,1,1,m_smallfont,command_line);
	
		for (int32 i=0;i<num_params;i++)
		{
			BString *string = m_string_stack->Pop();
			if (string->CountChars() == 0)
				menu->AddSeparator();
			else
				menu->AddItem(string->String(),ParseMenu(WarPoint(0,0),""));

			delete string;
		}
		return menu;
	}
	return NULL;
}

int32 WarGLView::execute(char *word,char *command_line,bool quiet,int32 call_stack)
{
//	printf("execute(%s,%s,%s,%ld)\n",word,command_line,quiet?"true":"false",call_stack);

	// the returned value, the number of characters to jump in the commandline befor the next
	// word is executed. Most commands wont touch this.
	int32 jump = 0;

	uint16 i=0;
	char *defined_word;
	bool is_word_defined;

	// should probably replace this with some search tree
	while (is_word_defined = m_word_list->GetNextItem(&defined_word,&i))
		if (!strcmp(word,defined_word)) break;
	i--;
	
	if (!is_word_defined) // the string we got is not a defined word
	{
		int32 ival;
		double fval;
		if (atoi(&ival,word))			// if it's an integer value, push to the integer stack
			m_integer_stack->Push(ival);
		else if (atof(&fval,word))		// if it's a float value, push it to the float stack
			m_float_stack->Push(fval);
		else if (!quiet) // or is it just another bad word
		{
			char tmp[256];
			sprintf(tmp,"\"%s\", unknown word",word);
			throw new WarException(tmp);
		}
		return 0;
	}

	if (i>=num_words) // the word is userdefined
	{
		char *command_line=new char[strlen(defined_word+strlen(defined_word)+1)+1];

		// the string contains the string to be executed after the word name
		strcpy(command_line,defined_word+strlen(defined_word)+1);
		
//		printf("userdefined word: '%s'\n",command_line);
	
		// execute the userdefined command line
		Parse(command_line,false,call_stack + 1);
		delete []command_line;
		return 0;
	}

	// the word is not userdefined
	switch (i)
	{



// constants
		case WORD_FALSE: // false
		{
			m_integer_stack->Push(0);
		}break;
		case WORD_TRUE: // true
		{
			m_integer_stack->Push(-1);
		}break;

// output functions
		case WORD_PRINT_STRING: // ."
		{
			int32 counter = 0;
			while (command_line[counter] != '"' && command_line[counter] != 0) counter++;

			jump = counter + 1;

			BString *string = new BString();

			string->Append(command_line,jump - 1);

			m_war_text_output->PutLine(string->String());
			delete string;
		}break;
		case WORD_TRACE_STACK: // .s
		{
			char *string = new char[m_integer_stack->CountElements() * 14 + 1];
			string[0] = 0;

			for (int32 i=0;i<m_integer_stack->CountElements();i++)
			{
				char tmp[14];
				sprintf(tmp,"%ld ",m_integer_stack->Pick(i));
				strcat(string,tmp);
			}

			m_war_text_output->PutLine(string);
			delete []string;
		}break;
		case WORD_POP: // .
		{
			int32 val;
			val = m_integer_stack->Pop();

			char tmp[10];
			sprintf(tmp,"%ld",val);
			m_war_text_output->PutLine(tmp);
		}break;
//	WORD_EMIT,
		case WORD_TYPE: // .t
		{
			BString *string = m_string_stack->Pop();
			m_war_text_output->PutLine(string->String());
			delete string;
		}break;
// WORD_CR,
//	WORD_SPACE,
//	WORD_SPACES,

// arithmetic operators
		case WORD_ADD: // +
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val1+val2);
		}break;
		case WORD_SUB: // -
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2-val1);
		}break;
		case WORD_MUL: // *
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val1*val2);
		}break;
		case WORD_DIV: // /
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2/val1);
		}break;
		case WORD_MODULO: // mod
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 % val1);
		}break;
		case WORD_DIV_MODULO: // /mod
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 / val1);
			m_integer_stack->Push(val2 % val1);
		}break;
		case WORD_INC: // inc
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push(val+1);
		}break;
		case WORD_DEC: // dec
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push(val-1);
		}break;
//	WORD_MAX,
//	WORD_MIN,
//	WORD_ABS,
//	WORD_NEGATE,
//	WORD_LEFT_SHIFT,
//	WORD_RIGHT_SHIFT,

// boolean operators
		case WORD_EQUALS: // =
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push( val2 == val1 ? -1 : 0);
		}break;
		case WORD_LESS: // <
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push( val2 < val1 ? -1 : 0);
		}break;
		case WORD_GREATER: // >
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push( val2 > val1 ? -1 : 0);
		}break;
		case WORD_NOT_EQUAL: // !=
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push( val2 != val1 ? -1 : 0);
		}break;
		case WORD_EQUAL_TO_ZERO: // 0=
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push( val == 0 ? -1 : 0);
		}break;
		case WORD_LESS_THAN_ZERO: // 0<
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push( val < 0 ? -1 : 0);
		}break;
		case WORD_GREATER_THAN_ZERO: // 0>
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push( val > 0 ? -1 : 0);
		}break;
		case WORD_AND: // and
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 && val1 ? -1 : 0);
		}break;
		case WORD_OR: // or
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 || val1 ? -1 : 0);
		}break;
		case WORD_NOT: // not
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Push(val ? 0 : -1);
		}break;

// bitwise operators
		case WORD_BITWISE_AND: // &
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 & val1);
		}break;
		case WORD_BITWISE_OR: // |
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 | val1);
		}break;
		case WORD_BITWISE_XOR: // ^
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2 ^ val1);
		}break;

// string stack manipulation
		case WORD_PUSH_STRING: // "
		{
			int32 counter = 0;
			while (command_line[counter] != '"' && command_line[counter] != 0) counter++;

			jump = counter + 1;

			BString *string = new BString();

			string->Append(command_line,jump - 1);

			m_string_stack->Push(string);
		}break;

// integer stack manipulation
		case WORD_DROP: // drop
		{
			m_integer_stack->Pop();
		}break;
		case WORD_2DROP: // 2drop
		{
			m_integer_stack->Pop();
			m_integer_stack->Pop();
		}break;
		case WORD_DUP: // dup
		{
			int32 val;
			val = m_integer_stack->Top();
			m_integer_stack->Push(val);
		}break;
		case WORD_NON_ZERO_DUP: // ?dup
		{
			int32 val;
			val = m_integer_stack->Top();
			if (val) m_integer_stack->Push(val);
		}break;
		case WORD_2DUP: // 2dup
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val2);
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val2);
			m_integer_stack->Push(val1);
			m_integer_stack->Empty();
		}break;
		case WORD_SWAP: // swap
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val2);
		}break;
		case WORD_2SWAP: // 2swap
		{
			int32 val1,val2,val3,val4;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			val3 = m_integer_stack->Pop();
			val4 = m_integer_stack->Pop();
			m_integer_stack->Push(val2);
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val4);
			m_integer_stack->Push(val3);
		}break;
		case WORD_ROTATE: // rot
		{
			int32 val1,val2,val3;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			val3 = m_integer_stack->Pop();
			m_integer_stack->Push(val2);
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val3);
		}break;
		case WORD_MINUS_ROTATE: // -rot
		{
			int32 val1,val2,val3;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			val3 = m_integer_stack->Pop();
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val3);
			m_integer_stack->Push(val2);
		}break;
		case WORD_NIP: // nip
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_integer_stack->Pop();
			m_integer_stack->Push(val);
		}break;
		case WORD_TUCK: // tuck
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pop();
			val2 = m_integer_stack->Pop();
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val2);
			m_integer_stack->Push(val1);
		}break;
		case WORD_OVER: // over
		{
			int32 val;
			val = m_integer_stack->Pick(1);
			m_integer_stack->Push(val);
		}break;
		case WORD_2OVER: // 2over
		{
			int32 val1,val2;
			val1 = m_integer_stack->Pick(3);
			val2 = m_integer_stack->Pick(2);
			m_integer_stack->Push(val1);
			m_integer_stack->Push(val2);
		}break;
//	WORD_PICK,
		case WORD_CLEAR: // clear
		{
			m_integer_stack->Empty();
		}break;

//
		case WORD_DEPTH: // depth
		{
			m_integer_stack->Push(m_integer_stack->CountElements());
		}break;
		case WORD_WORDS: // words
		{
			uint16 i=0;
			char *defined_word;
			while (m_word_list->GetNextItem(&defined_word,&i))
				m_war_text_output->PutLine(defined_word);
		}break;
		case WORD_EXEC: // execute
		{
			BString *string = m_string_stack->Pop();
			BFile file(m_script_dir,string->String(),B_READ_ONLY);
			
			if (file.InitCheck()==B_OK)
			{
				off_t len;
				file.GetSize(&len);
				// allocate a buffer to store the file
				char *buffer = new char[len + 1];
				
				// read the file
				file.Read(buffer,len);

				// replace all new-lines with spaces
				for (off_t i = 0;i<len;i++)
				{
					if (buffer[i] == '\n') buffer[i]=' ';
					else if (buffer[i] == '\t') buffer[i]=' ';
				}

				// add the NULL-termination
				buffer[len] = 0;
				
				// and parse the buffer
				Parse(buffer,quiet,call_stack + 1);
				delete []buffer;
			}
			else
			{
				char s[100 + B_FILE_NAME_LENGTH];
				sprintf(s,"file not found '%s'",string->String());
				delete string;
				throw new WarException(s);
			}
			delete string;
		}break;


// control structures
		case WORD_IF: // if
		{
			int32 val;
			val = m_integer_stack->Pop();
			if (val == 0)
			{
				int32 len = strlen(command_line);
				int32 nested = 1;
				char *cookie = NULL;
				while (nested > 0 && (word = GetWord(command_line,&cookie)))
				{
					// if we find a string, skip it
					if (!strcmp(word,words[WORD_PUSH_STRING]))
					{
						int32 counter = 0;
						while (cookie[counter] != '"' && cookie[counter] != 0) counter++;
						if (cookie[counter] == '"') counter++;
						cookie+= counter;
					}
					else if (!strcmp(word,words[WORD_START_COMMENT]))
						while ((word = GetWord(cookie,&cookie)) && strcmp(word,words[WORD_END_COMMENT]));
					else if (!strcmp(word,words[WORD_IF])) nested++;
					else if (!strcmp(word,words[WORD_ELSE])) nested--;
					else if (!strcmp(word,words[WORD_THEN])) nested--;
				}
				if (cookie == NULL) throw new WarException("'then' expected, use it to terminate an if-statement");
				else jump = len - strlen(cookie);
			}
		}break;
		case WORD_ELSE: // else
		{
			int32 len = strlen(command_line);
			int32 nested = 1;
			char *cookie = NULL;
			while (nested > 0 && (word = GetWord(command_line,&cookie)))
			{
				// if we find a string, skip it
				if (!strcmp(word,words[WORD_PUSH_STRING]))
				{
					int32 counter = 0;
					while (cookie[counter] != '"' && cookie[counter] != 0) counter++;
					if (cookie[counter] == '"') counter++;
					cookie+= counter;
				}
				else if (!strcmp(word,words[WORD_START_COMMENT]))
					while ((word = GetWord(cookie,&cookie)) && strcmp(word,words[WORD_END_COMMENT]));
				else if (!strcmp(word,words[WORD_IF])) nested++;
				else if (!strcmp(word,words[WORD_THEN])) nested--;
			}
			if (cookie == NULL) throw new WarException("'then' expected, use it to terminate an if-statement");
			else jump = len - strlen(cookie);
		}break;
		case WORD_THEN: // then
		{
		}break;
		case WORD_CASE:
		{
		}break;
		case WORD_ENDCASE:
		{
			m_integer_stack->Pop(); // drop
		}break;
		case WORD_OF:
		{
			int32 val1 = m_integer_stack->Pop();
			int32 val2 = m_integer_stack->Top();
			if (val1 != val2)
			{
				int32 len = strlen(command_line);
				int32 nested = 1;
				char *cookie = NULL;
				while (nested > 0 && (word = GetWord(command_line,&cookie)))
				{
					// if we find a string, skip it
					if (!strcmp(word,words[WORD_PUSH_STRING]))
					{
						int32 counter = 0;
						while (cookie[counter] != '"' && cookie[counter] != 0) counter++;
						if (cookie[counter] == '"') counter++;
						cookie+= counter;
					}
					else if (!strcmp(word,words[WORD_START_COMMENT]))
						while ((word = GetWord(cookie,&cookie)) && strcmp(word,words[WORD_END_COMMENT]));
					else if (!strcmp(word,words[WORD_OF])) nested++;
					else if (!strcmp(word,words[WORD_ENDOF])) nested--;
				}
				if (cookie == NULL) throw new WarException("'endof' expected, use it to terminate an of-statement");
				else jump = len - strlen(cookie);
			}
		}break;
		case WORD_ENDOF:
		{
			int32 len = strlen(command_line);
			int32 nested = 1;
			char *cookie = NULL;
			while (nested > 0 && (word = GetWord(command_line,&cookie)))
			{
				// if we find a string, skip it
				if (!strcmp(word,words[WORD_PUSH_STRING]))
				{
					int32 counter = 0;
					while (cookie[counter] != '"' && cookie[counter] != 0) counter++;
					if (cookie[counter] == '"') counter++;
					cookie+= counter;
				}
				else if (!strcmp(word,words[WORD_START_COMMENT]))
					while ((word = GetWord(cookie,&cookie)) && strcmp(word,words[WORD_END_COMMENT]));
				else if (!strcmp(word,words[WORD_CASE])) nested++;
				else if (!strcmp(word,words[WORD_ENDCASE])) nested--;
			}
			if (cookie == NULL) throw new WarException("'endcase' expected, use it to terminate a case-statement");
			else jump = len - strlen(cookie) - strlen(words[WORD_ENDCASE]) - 1;
		}break;


// word related
		case WORD_CONSTANT: // constant
		{
			// pops the constant value
			int32 val = m_integer_stack->Pop();

			// reads the constant name
			int32 len = strlen(command_line);
			char *cookie = NULL;
			word = GetWord(command_line,&cookie);

			if (cookie == NULL) jump = len;
			else jump = len - strlen(cookie);

			if (word == NULL) // we didn't find a word
			{
				m_war_text_output->PutLine("the constant word must be folowed by the constant name");
				break;
			}
			// defines a new word
			char *string = new char[jump + 10];
			sprintf(string,"%s %ld",word,val);

			uint16 i=0;
			while (string[i]!=' ' && string[i]!=0)
				i++;
			string[i]=0;

			// check if the word already is defined
			i=0;
			char *defined_word = NULL;
			while (m_word_list->GetNextItem(&defined_word,&i))
			{
				if (!strcmp(string,defined_word))
				{
					char s[256];
					sprintf(s,"Can't define the word \"%s\", it is already defined",string);
					throw new WarException(s);
					delete []string;
					string = NULL;
				}
			}

			if (string) m_word_list->AddItem(string);
		}break;
		case WORD_COMPILE: // :
		{
			char *line = strdup(command_line);

			int32 len = strlen(command_line);
			char *cookie = NULL;
			int32 nested = 1;
			while (nested > 0 && (word = GetWord(line,&cookie)))
			{
				if (!strcmp(word,words[WORD_PUSH_STRING]))
				{
					int32 counter = 0;
					while (cookie[counter] != '"' && cookie[counter] != 0) counter++;
					if (cookie[counter] == '"') counter++;
					cookie+= counter;
				}
				else if (!strcmp(word,words[WORD_START_COMMENT]))
					while ((word = GetWord(cookie,&cookie)) && strcmp(word,words[WORD_END_COMMENT]));
				else if (!strcmp(word,words[WORD_COMPILE])) nested++;
				else if (!strcmp(word,words[WORD_END_COMPILE])) nested--;
			}

			if (cookie == NULL) throw new WarException("';' expected, use it to end a function definition");
			else jump = len - strlen(cookie);

			char *string = new char[jump + 1];
			if (word == NULL) // we didn't find the end_compile word
			{
				strncpy(string,command_line,jump);
				string[jump] = 0;
			}
			else
			{
				strncpy(string,command_line,jump - strlen(words[WORD_END_COMPILE]) - 1);
				string[jump - strlen(words[WORD_END_COMPILE]) - 1] = 0;
			}

			// insert the NULL between the word-name and the function definition (the command line to be compiled)
			uint16 i=0;
			while (string[i]!=' ' && string[i]!=0)
				i++;
			string[i]=0;

			// check if the word already is defined
			i=0;
			char *defined_word = NULL;
			while (m_word_list->GetNextItem(&defined_word,&i))
			{
				if (!strcmp(string,defined_word))
				{
					char s[256];
					sprintf(s,"Can't compile, the word \"%s\" is already defined",string);
					m_war_text_output->PutLine(s);
					delete []string;
					string = NULL;
				}
			}
			if (string) m_word_list->AddItem(string);

			free(line);
		}break;
		case WORD_END_COMPILE: // ;
		{
			m_war_text_output->PutLine("you can't mark the end of a definition without starting one");
		}break;
		case WORD_FORGET: // forget
		{
			m_war_text_output->PutLine("forget can only be used at the beginning of a line");
		}break;

// comments
		case WORD_START_COMMENT: // (
		{
			int32 len = strlen(command_line);
			int32 nested = 1;
			char *cookie = NULL;
			while (nested > 0 && (word = GetWord(command_line,&cookie)))
			{
				if (!strcmp(word,words[WORD_START_COMMENT])) nested++;
				else if (!strcmp(word,words[WORD_END_COMMENT])) nested--;
			}
			if (cookie == NULL) throw new WarException("')' expected, use it to terminate a comment");
			else jump = len - strlen(cookie);
		}break;
		case WORD_END_COMMENT: // )
		{
		}break;
	
	
	
	// ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
	
	
/*
		case WORD_GAME_SET_MAP_MODE: // set_map_mode
		{
			int32 val;
			if (stack->Pop(&val))
				war_game->SetMapMode(val);
		}break;
		case WORD_GAME_MAP_MODE: // map_mode
		{
			stack->Push(war_game->MapMode());
		}break;
*/

//*******************************************************************************
//*******************************************************************************
//								GAME RELATED OPERATIONS
//*******************************************************************************
//*******************************************************************************

		case WORD_SET_CURSOR: // set_cursor
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_war_cli->SetCursorType(val);
		} break;		
		case WORD_CURSOR: // cursor
		{
			m_integer_stack->Push(m_war_cli->CursorType());
		} break;

		case WORD_GAME_MODIFIERS: // modifiers
		{
			m_integer_stack->Push(modifiers());
		}break;

		case WORD_GAME_CONSOLE_SHOW: // activate_console
		{
			int32 val = m_integer_stack->Pop();
			if (val && m_war_cli->IsHidden())
			{
				m_war_cli->Hide(false);
				GiveFocus(m_war_cli);
			}
			else if (!val && m_focus_stack->HasItem(m_war_cli))
			{
				m_war_cli->Hide(true);
				m_focus_stack->Remove(m_war_cli);
			}
		}break;
		case WORD_GAME_CONSOLE_IS_VISIBLE: // console_is_visible
		{
			if (m_war_cli->IsHidden())
				m_integer_stack->Push(0);
			else m_integer_stack->Push(-1);
		}break;

		case WORD_GAME_INPUT: // input
		{
			m_war_string_input->Hide(false);
			GiveFocus(m_war_string_input);
			m_war_string_input->SetCommandLine(command_line);
			
			// dont execute the rest of the command line yet
			// the war_string_input class will execute it when 
			// the user have typed a string
			jump = strlen(command_line);
		}break;

		case WORD_GAME_SET_FULLSCREEN_MODE: // set_fullscreen_mode
		{
			int32 val;
			val = m_integer_stack->Pop();
				((BDirectWindow*)Window())->SetFullScreen(val);
		}break;
		case WORD_GAME_FULLSCREEN_MODE: // fullscreen_mode
		{
			m_integer_stack->Push(((BDirectWindow*)Window())->IsFullScreen());
		}break;

		case WORD_GAME_STATS_SHOW: // stats.show
		{
			int32 val = m_integer_stack->Pop();
			m_war_statview->Hide(val == 0);
		}break;
		case WORD_GAME_STATS_IS_VISIBLE: // stats.is_visible
		{
			m_integer_stack->Push(m_war_statview->IsHidden()?0:-1);
		}break;

		case WORD_GAME_SAY: // say
		{
//			pop_string(tmp,256);
//			war_game->Say(tmp);
		}break;
		case WORD_GAME_QUIT: // quit
		{
			be_app->PostMessage(B_QUIT_REQUESTED);
		}break;
		case WORD_GAME_MENU: // menu
		{
			int32 x,y;
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();
		
			WarMenu *menu = ParseMenu(WarPoint(x,y),command_line);

			AddChild(menu);
			GiveFocus(menu);
			
			jump = strlen(command_line);
		}break;
		case WORD_GAME_MOUSE_POSITION: // mouse_position
		{
			WarPoint pos = m_war_mouse_view->MousePosition();
			m_integer_stack->Push(pos.x);
			m_integer_stack->Push(pos.y);
		}break;


//*******************************************************************************
//*******************************************************************************
//									SELECTION OPERATIONS
//*******************************************************************************
//*******************************************************************************

		case WORD_GAME_SELECTION_DESELECT_UNITS: // deselect_units
		{
			m_war_entity_list->DeselectAll(m_team);
		}break;
		case WORD_GAME_SELECTION_ASSIGN_GROUP: // assign_group
		{
			int32 group_id;
			group_id = m_integer_stack->Pop();

			if (group_id>254 || group_id < 0) m_war_text_output->PutLine("group id supposed to be in range 0-254");
			else m_war_entity_list->AssignGroup(m_team,group_id);
		}break;
		case WORD_GAME_SELECTION_SELECT_GROUP: // select_group
		{
			int32 group_id;
			group_id = m_integer_stack->Pop();

			if (group_id>254 || group_id < 0) m_war_text_output->PutLine("group supposed to be in range 0-254");
			else m_war_entity_list->SelectGroup(m_team,group_id);
		}break;
		case WORD_GAME_SELECTION_DESELECT_GROUP: // deselect_group
		{
			int32 group_id;
			group_id = m_integer_stack->Pop();

			if (group_id>254 || group_id < 0) m_war_text_output->PutLine("group supposed to be in range 0-254");
			else m_war_entity_list->DeselectGroup(m_team,group_id);
		}break;
		case WORD_GAME_SELECTION_LEAVE_GROUPS: // leave_groups
		{
			m_war_entity_list->LeaveGroups(m_team);
		}break;
		case WORD_GAME_SELECTION_COUNT_UNITS: // count units
		{
			m_integer_stack->Push(m_war_entity_list->SelectedEntities(m_team));
		}break;
		case WORD_GAME_SELECTION_UNIT_ID: // unit id
		{
			WarUnit *entity;
			uint16 id,cookie=0;
			if (m_war_entity_list->GetNextSelectedEntity(&entity,NULL,&id,&cookie)) m_integer_stack->Push(id);
			else m_integer_stack->Push(-1);
		}break;


//*******************************************************************************
//*******************************************************************************
//									CAMERA OPERATIONS
//*******************************************************************************
//*******************************************************************************

		case WORD_GAME_CAMERA_ATTACH: // camera.attach
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<0 || val>0xfffe) m_war_text_output->PutLine("The unit id have to be in the range 0 to 65534");
			else
			{
				WarUnit *body = m_war_entity_list->GetBodyByID(val);
				if (body != NULL) m_current_camera->AttachToUnit(body);
			}
		}break;
		case WORD_GAME_CAMERA_DETACH: // camera.detach
		{
			m_current_camera->DetachFromUnit();
		}break;
		case WORD_GAME_CAMERA_LOOK_AT_UNIT: // camera_look_at
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<0 || val>0xfffe) m_war_text_output->PutLine("The unit id have to be in the range 0 to 65534");
			else
			{
				WarUnit *body = m_war_entity_list->GetBodyByID(val);
				if (body != NULL) m_current_camera->LookAtUnit(body);
			}
		}break;
		case WORD_GAME_CAMERA_STOP_LOOK_AT_UNIT: // camera_stop_look_at
		{
			m_current_camera->StopLookAtUnit();
		}break;
		case WORD_GAME_CAMERA_LOCK: // camera_lock
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<0 || val>0xfffe) m_war_text_output->PutLine("The unit id have to be in the range 0 to 65534");
			else
			{
				WarUnit *body = m_war_entity_list->GetBodyByID(val);
				if (body != NULL) m_current_camera->LockToUnit(body);
			}
		}break;
		case WORD_GAME_CAMERA_UNLOCK: // camera_unlock
		{
			m_current_camera->UnlockFromUnit();
		}break;
		case WORD_GAME_CAMERA_ADD_MOVE_SPEED: // camera_add_move_speed
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->AddMovementSpeed(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_ADD_ROTATION_SPEED: // camera_add_rotation_speed
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->AddRotationSpeed(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_ADD_SCROLL_SPEED: // camera_add_scroll_scroll_speed
		{
			int32 vert,horiz;
			vert = m_integer_stack->Pop();
			horiz = m_integer_stack->Pop();
			m_current_camera->AddScrollSpeed(horiz,vert);
		}break;
		case WORD_GAME_CAMERA_ROTATE_BY: // camera_rotate_by
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->RotateBy(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_MOVE_BY: // camera_move_by
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->MoveBy(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_SET_FOV: // set fov
		{
			int32 fov;
			fov = m_integer_stack->Pop();

			if (fov<1 || fov>189) m_war_text_output->PutLine("The field of vision must be a number from 1 to 189");
			else m_current_camera->SetFov(fov/180.0 * 3.1415927,atan2(m_war_display->Height(),m_war_display->Width() / tan(fov/180.0 * 3.1415927)));
		}break;
		case WORD_GAME_CAMERA_FOV: // fov
		{
			m_integer_stack->Push(m_current_camera->VerticalFov() * 180.0 / 3.1415927);
		}break;
		case WORD_GAME_CAMERA_SET_FAR_CLIP: // set_far_clip
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<80) m_war_text_output->PutLine("The far z clip has to be at least 80.0");
			m_current_camera->SetFarClip(val);
		}break;
		case WORD_GAME_CAMERA_FAR_CLIP: // far_clip
		{
			m_integer_stack->Push(m_current_camera->FarClip());
		}break;
		case WORD_GAME_CAMERA_MOVE: // camera_move
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->MoveTo(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_LOOK_AT: // camera_look_at
		{
			int32 x,y,z;
			z = m_integer_stack->Pop();
			y = m_integer_stack->Pop();
			x = m_integer_stack->Pop();

			m_current_camera->LookAt(WarVertex(x,y,z));
		}break;
		case WORD_GAME_CAMERA_POSITION: // camera_position
		{
			WarVertex vertex = m_current_camera->Position();

			m_integer_stack->Push((int32)vertex.x);
			m_integer_stack->Push((int32)vertex.y);
			m_integer_stack->Push((int32)vertex.z);
		}break;


//*******************************************************************************
//*******************************************************************************
//									DISLPLAY OPERATIONS
//*******************************************************************************
//*******************************************************************************

		case WORD_GAME_DISPLAY_SWITCH_CAMERA: // display.switch_camera
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<0 || val>15) m_war_text_output->PutLine("The camera index is supposed to be in range 0 to 15");
			else
			{
				m_current_camera->Stop();
				m_current_camera = m_camera + val;
				m_war_display->SetCamera(m_current_camera);
			}
		}break;
		case WORD_GAME_DISPLAY_BEGIN_MOUSE_LOOK: // begin mouse look
		{
			m_war_display->BeginMouseLook();
		}break;
		case WORD_GAME_DISPLAY_END_MOUSE_LOOK: // end mouse look
		{
			m_war_display->EndMouseLook();
		}break;
		case WORD_GAME_DISPLAY_MOUSE_LOOK: // mouse look
		{
			m_integer_stack->Push(m_war_display->MouseLook());
		}break;
		case WORD_GAME_DISPLAY_BEGIN_MOUSE_SELECTION: // begin mouse selection
		{
			m_war_display->BeginMouseSelection();
		}break;
		case WORD_GAME_DISPLAY_END_MOUSE_SELECTION: // end mouse selection
		{
			m_war_display->EndMouseSelection();
		}break;
		case WORD_GAME_DISPLAY_MOUSE_SELECTION: // mouse selection
		{
			m_integer_stack->Push(m_war_display->MouseSelection());
		}break;
		case WORD_GAME_DISPLAY_BEGIN_MOUSE_SCROLL: // begin mouse scroll
		{
			m_war_display->BeginMouseScroll();
		}break;
		case WORD_GAME_DISPLAY_END_MOUSE_SCROLL: // end mouse scroll
		{
			m_war_display->EndMouseScroll();
		}break;
		case WORD_GAME_DISPLAY_MOUSE_SCROLL: // mouse scroll
		{
			m_integer_stack->Push(m_war_display->MouseScroll());
		}break;
		case WORD_GAME_DISPLAY_SET_DETAIL_LEVEL: // set_detail_level
		{
			int32 val;
			val = m_integer_stack->Pop();

			if (val<1)
			{
				m_war_text_output->PutLine("The detail level has to be at least 1");
				val = 1;
			}
			if (val>500)
			{
				m_war_text_output->PutLine("The detail level can't be greater than 500");
				val = 500;
			}
			m_war_display->SetDetailLevel(val);
		}break;
		case WORD_GAME_DISPLAY_DETAIL_LEVEL: // detail_level
		{
			m_integer_stack->Push(m_war_display->DetailLevel());
		}break;
		case WORD_GAME_DISPLAY_SET_WIREFRAME:
		{
			int32 val;
			val = m_integer_stack->Pop();
			m_war_display->SetWireframe(val);
		}break;
		case WORD_GAME_DISPLAY_WIREFRAME:
		{
			m_integer_stack->Push(m_war_display->IsWireframe()?1:0);
		}break;
	}

//	printf("returning: %ld\n",jump);
	return jump;
}

// the cookie should be initialized to NULL the first time the function call is made.
char *WarGLView::GetWord(char *command_line,char **cookie)
{
	char *line;
	if (*cookie==NULL) line=command_line;
	else line=*cookie;

	uint16 i=0;

	// removes spaces in the begining of the string
	while (*line == ' ' && *line != 0) line++;
/*
	if (line[i]=='"')
	{
		i++;
		while (line[i]!=0 && line[i]!='"')
			i++;
		if (line[i]!='"') m_war_text_output->PutLine("ERROR: expected \"");
		i++;
	}
	else */
	if (line[0]==0)
	{
		*cookie=NULL;
		return NULL;
	}
	else
		while (line[i]!=0 && line[i]!=' ')
			i++;
//	printf("line2: %s i:%d\n",line,i);

	if (line[i]!=0)
	{
		line[i]=0;
		i++;
	}
	*cookie=line+i;
	return line;
}

// this is my own version of the function atoi() (ascii to integer)
// this can see if the given string really contains a number and
// it also handles negative numbers
bool WarGLView::atoi(int32 *val,char *string)
{
	return sscanf(string,"%ld",val) == 1;
/*
	uint16 len=strlen(string);
	bool negative=false;
	int32 tmp;

	*val=0;

	if (len==0) return false;

	if (string[0]=='-') // if negative
	{
		negative=true;
		len--;
		string++;
	} else if (string[0]=='+') // if positive
	{
		negative=false;
		len--;
		string++;
	}

	uint16 i=0;
	while (i<len)
	{
		if (string[i]<'0' || string[i]>'9') return false;

		tmp=1;
		uint16 j;
		for (j=0;j<len-i-1;j++) // powers ten to the position value
			tmp*=10;

		*val+=tmp*(string[i]-'0');
		i++;
	}
	if (negative) *val=-*val;
	return true;
*/
}

bool WarGLView::atof(double *val,char *string)
{
	return sscanf(string,"%lf",val) == 1;
}
