#include "WarObject.h"
#include "WarFace.h"
#include "WarVertex.h"
#include "WarIFFChunk.h"
#include "WarCamera.h"

#include <math.h>
#include <gl.h>
#include <stdio.h>

_EXPORT WarObject::WarObject(WarObject *object_to_copy):WarMatrix()
{
// copy the sub-object list
	sub_objects = new WarList<WarObject*>(object_to_copy->ObjectCount(),1);
	for (uint16 i=0;i<object_to_copy->ObjectCount();i++)
		sub_objects->AddItem(new WarObject(object_to_copy->Object(i)));

// copy the objects model_id
	model_id=object_to_copy->model_id;

// copy the scale
	ScaleTo(object_to_copy->Scale());

// copy the angle
	RotateTo(object_to_copy->Direction());

// copy the position
	MoveTo(object_to_copy->Position());
}

_EXPORT WarObject::WarObject(BDirectory *dir,const char *filename):WarMatrix()
{
	sub_objects = new WarList<WarObject*>(1,1);

// init the scale
	ScaleTo(WarVertex(1.0,1.0,1.0));

// init the angle
	RotateTo(WarVertex(0.0,0.0,0.0));

// init the position
	MoveTo(WarVertex(0.0,0.0,0.0));


// open the file
	WarIFFChunk file(dir,filename,B_READ_ONLY);
	if (file.InitCheck() != B_OK)
	{
		fprintf(stderr,"error loading lightwave object: \"%s\"\n",filename);
		return;
	}
	
//	************* VECTORS

	WarIFFChunk points;
	off_t size;
	if (file.FindChunk('PNTS',&points) == B_ERROR) return;
	points.GetSize(&size);
	int32 vertex_count = size/12;

	WarVertex *vertices = new WarVertex[vertex_count];
	WarVertex *normals = new WarVertex[vertex_count];

	for (int c=0;c<vertex_count;c++)
	{
		points.ReadFloat(&vertices[c].x);
		points.ReadFloat(&vertices[c].y);
		points.ReadFloat(&vertices[c].z);
	}

//	************** POLYGONS

	WarIFFChunk polys;
	if (file.FindChunk('POLS',&polys) == B_ERROR) return;

//	fprintf(stderr,"0x%x size == %d faces\n",size,size/10);
	polys.GetSize(&size);
	WarFace **face=new WarFace*[size / 6];
	uint16 face_count = 0;
	
	uint16 tmp;

	uint32 read_bytes = 0;
	uint16 i = 0;

	while (size > read_bytes)
	{
		read_bytes += polys.ReadUint16(&tmp);
		face[i] = new WarFace(tmp);
		face_count++;

		for (uint16 j=0;j<face[i]->VertexCount();j++)
		{
			read_bytes += polys.ReadUint16(&tmp);
			face[i]->vex[j] = vertices + tmp;
		}

		read_bytes += polys.ReadUint16(&face[i]->material); // material
		i++;
	}

//	set normals
	for (int c=0;c<face_count;c++)
		face[c]->CalcNormal();

	WarVertex normal;
	uint16 num_faces;
	for (int i=0;i<vertex_count;i++)
	{
		num_faces=0;
		normal.SetTo(0,0,0);
		for (int c=0;c<face_count;c++)
			if (face[c]->vex[0]==&vertices[i] || face[c]->vex[1]==&vertices[i] || face[c]->vex[2]==&vertices[i])
			{
				normal += (WarVertex)face[c]->Normal();
				num_faces++;
			}
		if (num_faces) normal/=num_faces;
		normals[i]=normal;
	}

//	CalcBoundingBox();

// allocate a display_list id#
	model_id = glGenLists(1);
// if that fails, return
	if (model_id == 0) return;
// begin to draw the object
	glNewList(model_id,GL_COMPILE);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	glColor3f(.4,1.0,.1);
	for(uint16 i=0;i<face_count;i++)
	{
		if (face[i]->VertexCount() == 2)
			glBegin(GL_LINES);
		else if (face[i]->VertexCount() == 3)
			glBegin(GL_TRIANGLES);
		else if (face[i]->VertexCount() == 4)
			glBegin(GL_QUADS);
		else
			glBegin(GL_POLYGON);
		
		for (uint16 j=0;j<face[i]->VertexCount();j++)
		{
			glNormal3fv((GLfloat *)(normals + (face[i]->vex[j] - vertices)));
			glVertex3fv((GLfloat *)face[i]->vex[j]);
		}

		glEnd();

	}
	glEndList();
	

	for (int32 i=0;i<face_count;i++)
		delete face[i];

	delete face;
	delete vertices;
	delete normals;
}

_EXPORT void WarObject::AddSubObject(WarObject *obj)
{
	sub_objects->AddItem(obj);
}

_EXPORT void WarObject::Draw(WarCamera *camera)
{
	glPushMatrix();

	glTranslatef(XPosition(),YPosition(),ZPosition());


	glRotatef(ZDirection() * 180.0 / 3.1415927,0.0,0.0,1.0); // direction
	glRotatef(YDirection() * 180.0 / 3.1415927,0.0,1.0,0.0); // up-down
	glRotatef(XDirection() * 180.0 / 3.1415927,1.0,0.0,0.0); // tilt

	glScalef(XScale(),YScale(),ZScale());

	glCallList(model_id);
// draw the sub-objects
	for (uint16 i=0;i<sub_objects->CountItems();i++) sub_objects->ItemAt(i)->Draw(camera);
	glPopMatrix();
}

_EXPORT WarObject::~WarObject()
{
// delete all the sub-objects
	for (uint16 i=0;i<sub_objects->CountItems();i++)
		delete sub_objects->ItemAt(i);

// delete the object list
	delete sub_objects;
}

_EXPORT void WarObject::RotateToNormal(WarVertex *normal)
{
	float za = ZDirection();

	float y = normal->x*cos(za) + normal->y*sin(za);
	float x = normal->x*sin(za) - normal->y*cos(za);

	RotateXTo(atan(x/normal->z));
	RotateYTo(atan(y/normal->z));

//	float x=normal->x;
//	float y=normal->y;

//	angle_x=atan(-x/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(-y/normal->z)*360.0/(2*3.1415927); // z:0

//	angle_x=atan(y/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(-x/normal->z)*360.0/(2*3.1415927); // z:90

//	angle_x=atan(x/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(y/normal->z)*360.0/(2*3.1415927); // z:180

//	angle_x=atan(-y/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(x/normal->z)*360.0/(2*3.1415927); // z:270

}
