#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <Rect.h>

#include "WarDisplay.h"
#include "WarTriangulation.h"
#include "Global.h"

_EXPORT WarTriangulation::WarTriangulation(WarMap *map)
{
	war_map = map;
	quad_tree = new uint8[(war_map->Width()+1) * (war_map->Height()+1)];
	memset(quad_tree,0,(war_map->Width()+1) * (war_map->Height()+1));
	
	box = new WarBox();

#ifdef _MEMORY_STATISTICS_
	fprintf(stderr,"WarTriangulation\n");
	fprintf(stderr,"  Quad-tree: %d\n",(war_map->Width()+1) * (war_map->Height()+1));
#endif

	queue = new WarHeapf(war_map->Width() * war_map->Height());
	
	detail_level = 4.0;
}

_EXPORT WarTriangulation::~WarTriangulation()
{
	delete queue;
	delete box;
	delete []quad_tree;
}

_EXPORT void WarTriangulation::Regenerate(WarCamera *war_camera)
{

	queue->Empty();

	memset(quad_tree,0,(war_map->Width()+1) * (war_map->Height()+1));

	generate_map_triangulation(war_map->Width()/2,war_map->Height()/2,war_map->Width()/2,war_camera);
	uint16 x,y;
	uint8 level;
	float weight;

	while (queue->Pop(&x,&y,&level,&weight))
		generate_tile_triangulation(x,y,level,weight,war_camera);
}

_EXPORT void WarTriangulation::GenerateStatic()
{
	queue->Empty();
	memset(quad_tree,0,(war_map->Width()+1) * (war_map->Height()+1));

	generate_map_triangulation(war_map->Width()/2,war_map->Height()/2,war_map->Width()/2);
	uint16 x,y;
	uint8 level;
	float weight;

	while (queue->Pop(&x,&y,&level,&weight))
		generate_tile_triangulation(x,y,level,weight);

}

void WarTriangulation::generate_map_triangulation(int32 x,int32 y,int32 level,WarCamera *camera)
{
	box->SetTo(level * 2,level * 2,64.0);
	box->OffsetBy(WarVertex(x,y,32.0));

	if (camera->IsBoxVisible(box))
	{
		if (level == TILE_SPAN>>1)
		{
			float weight;
			if ((weight = subdivide(x,y,level,camera)) < 1.0)
//				generate_tile_triangulation(x,y,level,weight,camera);
				queue->Push(x,y,level,weight);
			else
				quad_tree[(y<<war_map->WidthShift()) + y + x] = 0;
		}
		else
		{
			int32 half_level = level>>1;
			generate_map_triangulation(x+half_level,y+half_level,half_level,camera);
			generate_map_triangulation(x+half_level,y-half_level,half_level,camera);
			generate_map_triangulation(x-half_level,y-half_level,half_level,camera);
			generate_map_triangulation(x-half_level,y+half_level,half_level,camera);
		}
	}
	else
		quad_tree[ (y<<war_map->WidthShift()) + y + x] = 255;
}

float WarTriangulation::subdivide(int32 x,int32 y, int32 level,WarCamera *camera)
{
	if (x < 0 || y < 0 || x > war_map->Width() || y > war_map->Height()) return 1.0;
	#define p(x) ((x)*(x))

	float distance = p(war_map->GetHeight(x,y) - camera->ZPosition()) + p(x - camera->XPosition()) + p(y - camera->YPosition());

	float c = detail_level * (float)war_map->Bump(x,y)/8.0;
	if (c < 1.0) c = 1.0;

//	return 0.0;
//	printf("%f\n",1000.0/( p((float)level) * p(20.0) * p(c) ));
//	return (war_map->Bump(x,y)?0.0:1.0);

	return distance/( p((float)level) * p(20.0) * c);
}

void WarTriangulation::generate_tile_triangulation(int32 x,int32 y,int32 level,float blend,WarCamera *camera)
{
	// level describes the radius (half the lenght of a side) of the square
	if (level == TILE_SPAN) return;

	int8 half_level = level>>1;

	static int32 sinus1[]={1,1,-1,-1};
	static int32 cosin1[]={1,-1,-1,1};

	// we have to check that the resolution differ no more than one level, if
	// it does, force split
	for (int32 j = 0;j<4;j++)
	{
		int32 x_pos = x % (4*level) - 2*level;
		int32 y_pos = y % (4*level) - 2*level;

		int32 tmp_x = x + 2*level*cosin1[j] + x_pos;
		int32 tmp_y = y + 2*level*sinus1[j] + y_pos;
		if (Lookup(tmp_x,tmp_y) == 0)
		{
//			if (tmp_x <0 || tmp_x > war_map->Width() || tmp_y < 0 || tmp_y > war_map->Height()) printf("AHH!!! 2\n");
//			return;
			generate_tile_triangulation(tmp_x,tmp_y,2*level,0.0,camera);
		}
	}



	if (blend>1.0) blend = 1.0;
	else if (blend<0.5) blend = 0.5;

//	uint8 blend_val = (uint8)(2.0 * (1.0 - blend) * 255.0);
//	quad_tree[y * (war_map->Width() + 1) + x] = blend;
//	quad_tree[(y - level) * (war_map->Width() + 1) + (x + level)] = blend;
//	quad_tree[(y + level) * (war_map->Width() + 1) + (x - level)] = blend;
//	quad_tree[(y + level) * (war_map->Width() + 1) + (x + level)] = blend;
//	quad_tree[(y - level) * (war_map->Width() + 1) + (x - level)] = blend;

	set(x,y,255);
	set(x+level,y-level,255);
	set(x-level,y+level,255);
	set(x+level,y+level,255);
	set(x-level,y-level,255);

	// we have rached the most detailed level,
	// we cannot go any further down the tree
	if (level == 1) return;
	
	for (int32 i=0;i<4;i++)
	{
		int32 x_add1 = half_level * cosin1[i];
		int32 y_add1 = half_level * sinus1[i];

		set(x + x_add1, y + y_add1,0);

		float weight;

//		if (x + x_add1 <0 || x + x_add1 > war_map->Width() || y + y_add1 < 0 || y + y_add1 > war_map->Height()) printf("AHH!!! 3\n");

		if ((weight = subdivide(x + x_add1,y + y_add1,half_level,camera)) < 1.0)
		{
			queue->Push(x + x_add1,y + y_add1,half_level,weight);
//			generate_tile_triangulation(x + x_add1,y + y_add1,half_level,weight,camera);
		}
	}
}









void WarTriangulation::generate_map_triangulation(int32 x,int32 y,int32 level)
{
	if (level == TILE_SPAN>>1)
	{
		float weight;
		if ((weight = subdivide(x,y,level)) < 1.0)
//			generate_tile_triangulation(x,y,level,weight,camera);
			queue->Push(x,y,level,weight);
		else
			quad_tree[(y<<war_map->WidthShift()) + y + x] = 0;
	}
	else
	{
		int32 half_level = level>>1;
		generate_map_triangulation(x+half_level,y+half_level,half_level);
		generate_map_triangulation(x+half_level,y-half_level,half_level);
		generate_map_triangulation(x-half_level,y-half_level,half_level);
		generate_map_triangulation(x-half_level,y+half_level,half_level);
	}
}

float WarTriangulation::subdivide(int32 x,int32 y, int32 level)
{
	if (x < 0 || y < 0 || x > war_map->Width() || y > war_map->Height()) return 1.0;
	#define p(x) ((x)*(x))

//	printf("%f\n",((float)(255 - war_map->Bump(x,y))));

	float c = ((float)(255 - war_map->Bump(x,y)))/245.0;

	return c;
}

void WarTriangulation::generate_tile_triangulation(int32 x,int32 y,int32 level,float blend)
{
	// level describes the radius (half the lenght of a side) of the square
	if (level == TILE_SPAN) return;

	int8 half_level = level>>1;

	static int32 sinus1[]={1,1,-1,-1};
	static int32 cosin1[]={1,-1,-1,1};

	// we have to check that the resolution differ no more than one level, if
	// it does, force split
	for (int32 j = 0;j<4;j++)
	{
		int32 x_pos = x % (4*level) - 2*level;
		int32 y_pos = y % (4*level) - 2*level;

		int32 tmp_x = x + 2*level*cosin1[j] + x_pos;
		int32 tmp_y = y + 2*level*sinus1[j] + y_pos;
		if (Lookup(tmp_x,tmp_y) == 0)
		{
//			if (tmp_x <0 || tmp_x > war_map->Width() || tmp_y < 0 || tmp_y > war_map->Height()) printf("AHH!!! 2\n");
//			return;
			generate_tile_triangulation(tmp_x,tmp_y,2*level,0.0);
		}
	}



	if (blend>1.0) blend = 1.0;
	else if (blend<0.5) blend = 0.5;

//	uint8 blend_val = (uint8)(2.0 * (1.0 - blend) * 255.0);
//	quad_tree[y * (war_map->Width() + 1) + x] = blend;
//	quad_tree[(y - level) * (war_map->Width() + 1) + (x + level)] = blend;
//	quad_tree[(y + level) * (war_map->Width() + 1) + (x - level)] = blend;
//	quad_tree[(y + level) * (war_map->Width() + 1) + (x + level)] = blend;
//	quad_tree[(y - level) * (war_map->Width() + 1) + (x - level)] = blend;

	set(x,y,255);
	set(x+level,y-level,255);
	set(x-level,y+level,255);
	set(x+level,y+level,255);
	set(x-level,y-level,255);

	// we have rached the most detailed level,
	// we cannot go any further down the tree
	if (level == 1) return;
	
	for (int32 i=0;i<4;i++)
	{
		int32 x_add1 = half_level * cosin1[i];
		int32 y_add1 = half_level * sinus1[i];

		set(x + x_add1, y + y_add1,0);

		float weight;

//		if (x + x_add1 <0 || x + x_add1 > war_map->Width() || y + y_add1 < 0 || y + y_add1 > war_map->Height()) printf("AHH!!! 3\n");

		if ((weight = subdivide(x + x_add1,y + y_add1,half_level)) < 1.0)
		{
			queue->Push(x + x_add1,y + y_add1,half_level,weight);
//			generate_tile_triangulation(x + x_add1,y + y_add1,half_level,weight,camera);
		}
	}
}
