#include "WarObject.h"
#include "WarFace.h"
#include "WarVertex.h"
#include "WarIFFFile.h"

#include <math.h>
#include <gl.h>
#include <stdio.h>

_EXPORT WarObject::WarObject(WarObject *object_to_copy):WarMatrix()
{
// copy the number of sub-objects
	object_count=object_to_copy->object_count;
// copy the objects model_id
	model_id=object_to_copy->model_id;

// copy the scale
	ScaleTo(object_to_copy->Scale());

// copy the angle
	RotateTo(object_to_copy->Direction());

// copy the position
	MoveTo(object_to_copy->Position());

// if there's any sub-objects, the nwe should copy them too
	if (object_count)
	{
// allocate the sub-object list
		sub_objects=new WarObject*[object_count];
// copy the sub-objects recursively
		for (uint16 i=0;i<object_count;i++) sub_objects[i]=new WarObject(object_to_copy->sub_objects[i]);
	}
}

_EXPORT WarObject::WarObject(BDirectory *dir,const char *filename):WarMatrix()
{
// init the scale
	ScaleTo(WarVertex(1.0,1.0,1.0));

// init the angle
	RotateTo(WarVertex(0.0,0.0,0.0));

// init the position
	MoveTo(WarVertex(0.0,0.0,0.0));

// initialize the sub_object list
	sub_objects = NULL;
	object_count = 0;

// open the file
	WarIFFFile file(dir,filename,B_READ_ONLY);
	if (file.InitCheck() != B_OK)
	{
		fprintf(stderr,"error loading lightwave object: \"%s\"\n",filename);
		return;
	}
	
//	************* VECTORS

	uint32 size;
	if (file.FindChunk('PNTS',&size) == B_ERROR) return;
	int32 vertex_count = size/12;

	WarVertex *vertices = new WarVertex[vertex_count];
	WarVertex *normals = new WarVertex[vertex_count];

	for (int c=0;c<vertex_count;c++)
	{
		file.ReadFloat(&vertices[c].x);
		file.ReadFloat(&vertices[c].y);
		file.ReadFloat(&vertices[c].z);
	}

//	************** POLYGONS

	if (file.FindChunk('POLS',&size) == B_ERROR) return;

//	fprintf(stderr,"0x%x size == %d faces\n",size,size/10);

	WarFace **face=new WarFace*[size / 6];
	uint16 face_count = 0;
	
	
	uint16 tmp;

	uint32 read_bytes = 0;
	uint16 i = 0;

	while (size > read_bytes)
	{
		read_bytes += file.ReadUint16(&tmp);
		face[i] = new WarFace(tmp);
		face_count++;

		for (uint16 j=0;j<face[i]->VertexCount();j++)
		{
			read_bytes += file.ReadUint16(&tmp);
			face[i]->vex[j] = vertices + tmp;
		}

		read_bytes += file.ReadUint16(&face[i]->material); // material
		i++;
	}

//	set normals
	for (int c=0;c<face_count;c++)
		face[c]->CalcNormal();

	WarVertex normal;
	uint16 num_faces;
	for (int i=0;i<vertex_count;i++)
	{
		num_faces=0;
		normal.SetTo(0,0,0);
		for (int c=0;c<face_count;c++)
			if (face[c]->vex[0]==&vertices[i] || face[c]->vex[1]==&vertices[i] || face[c]->vex[2]==&vertices[i])
			{
				normal += (WarVertex)face[c]->Normal();
				num_faces++;
			}
		if (num_faces) normal/=num_faces;
		normals[i]=normal;
	}

//	CalcBoundingBox();

// allocate a display_list id#
	model_id = glGenLists(1);
// if that fails, return
	if (model_id == 0) return;
// begin to draw the object
	glNewList(model_id,GL_COMPILE);
	glDisable(GL_TEXTURE_2D);
	for(uint16 i=0;i<face_count;i++)
	{
		if (face[i]->VertexCount() == 2)
			glBegin(GL_LINES);
		else if (face[i]->VertexCount() == 3)
			glBegin(GL_TRIANGLES);
		else if (face[i]->VertexCount() == 4)
			glBegin(GL_QUADS);
		else
			glBegin(GL_POLYGON);
		
		for (uint16 j=0;j<face[i]->VertexCount();j++)
		{
			glNormal3fv((GLfloat *)(normals + (face[i]->vex[j] - vertices)));
			glVertex3fv((GLfloat *)face[i]->vex[j]);
		}

		glEnd();

	}
	glEndList();
}

_EXPORT void WarObject::AddSubObject(WarObject *obj)
{
// allocate a new object list
	WarObject **new_object_list=new WarObject*[object_count+1];
// copy the old list to the new
	for (uint16 i=0;i<object_count;i++) new_object_list[i]=sub_objects[i];
// set the new last element
	new_object_list[object_count]=obj;
// increase the object counter by one
	object_count++;
// replace the old list with the new one
	delete []sub_objects;
	sub_objects=new_object_list;
}

_EXPORT void WarObject::Draw()
{
	glPushMatrix();

	glTranslatef(XPosition(),YPosition(),ZPosition());


	glRotatef(ZDirection() * 180.0 / 3.1415927,0.0,0.0,1.0); // direction
	glRotatef(YDirection() * 180.0 / 3.1415927,0.0,1.0,0.0); // up-down
	glRotatef(XDirection() * 180.0 / 3.1415927,1.0,0.0,0.0); // tilt

	glScalef(XScale(),YScale(),ZScale());

	glCallList(model_id);
// draw the sub-objects
	for (uint16 i=0;i<object_count;i++) sub_objects[i]->Draw();
	glPopMatrix();
}

_EXPORT WarObject::~WarObject()
{
// delete all the sub-objects
	for (uint16 i=0;i<object_count;i++) delete sub_objects[i];
// delete the object list
	delete []sub_objects;
}

_EXPORT void WarObject::RotateToNormal(WarVertex *normal)
{

	float za = ZDirection();

	float y = normal->x*cos(za) + normal->y*sin(za);
	float x = normal->x*sin(za) - normal->y*cos(za);

	RotateXTo(atan(x/normal->z));
	RotateYTo(atan(y/normal->z));

//	float x=normal->x;
//	float y=normal->y;

//	angle_x=atan(-x/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(-y/normal->z)*360.0/(2*3.1415927); // z:0

//	angle_x=atan(y/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(-x/normal->z)*360.0/(2*3.1415927); // z:90

//	angle_x=atan(x/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(y/normal->z)*360.0/(2*3.1415927); // z:180

//	angle_x=atan(-y/normal->z)*360.0/(2*3.1415927);
//	angle_y=atan(x/normal->z)*360.0/(2*3.1415927); // z:270

}
