/******************************************************
* cpp2html written by Arvid Norberg <arvid@iname.com> 2000
*****************************************************/

%option noyywrap
%{
	enum
	{
		NORMAL_MODE = 0,
		STRING_MODE,
		CHAR_MODE,
		COMMENT_MODE,
		SINGLELINE_COMMENT
	} mode = NORMAL_MODE;

	void echo(ostream &out, const char *pre_string, const char *string, const char *post_string);
%}

alpha		[A-Za-z]
dig			[0-9]
name		({alpha}|{dig}|\$)({alpha}|{dig}|[_.\-/$])*
num1		[-+]?{dig}+\.?([eE][-+]?{dig}+)?
num2		[-+]?{dig}*\.{dig}+([eE][-+]?{dig}+)?
num3		"0x"[0-9a-fA-F]+

modifiers		volatile|extern|register|mutable|static|virtual|private|protected|public|inline|const|auto|explicit|template|using|friend
controls		if|else|for|while|do|goto|continue|return|switch|case|break|default
definitions		class|struct|union|enum|namespace|typedef
operators		operator|or|not|xor|new|delete|sizeof|dynamic_cast|static_cast|const_cast|reinterpret_cast

number			{num1}|{num2}|{num3}
keyword			false|true|asm|typeid|{modifiers}|{controls}|{definitions}|{operators}
preprocessor	"#"[ |\t]*(if|define|ifndef|ifdef|else|include|endif|elif|line|undef|pragma|error)
datatype		int|char|short|long|void|bool|double|float|wchar_t|signed|unsigned
usertype		int8|uint8|vint8|vuint8|int16|uint16|vint16|vuint16|int32|uint32|vint32|vuint32|int64|uint64|vint64|vuint64|vlong|vint|vshort|vchar|vulong|vuint|vushort|vuchar|uchar|unichar

%%
"//" {
	if (mode == NORMAL_MODE)
	{
		*yyout << "<code class = \"comment\">";
		mode = SINGLELINE_COMMENT;
	}
	*yyout << YYText();
}

"\n" {
	if (mode == SINGLELINE_COMMENT)
	{
		*yyout << "</code>";
		mode = NORMAL_MODE;
	}
	*yyout << YYText();
}

"/*" {
	if (mode == NORMAL_MODE)
	{
		*yyout << "<code class = \"comment\">";
		mode = COMMENT_MODE;
	}
	*yyout << YYText();
}

"*/" {
	*yyout << YYText();
	if (mode == COMMENT_MODE)
	{
		*yyout << "</code>";
		mode = NORMAL_MODE;
	}
}

"<"	*yyout << "&lt;";
">"	*yyout << "&gt;";
"&"	*yyout << "&amp;";
"°"	*yyout << "&deg;";
"®"	*yyout << "&reg;";
"©"	*yyout << "&copy;";
"¤"	*yyout << "&curren;";
"ç"	*yyout << "&cent;";
"£"	*yyout << "&pound;";
"×"	*yyout << "&times;";
"÷"	*yyout << "&divide;";
"½"	*yyout << "&frac12;";
"¼"	*yyout << "&frac14;";
"¾"	*yyout << "&frac34;";
"¹"	*yyout << "&sup1;";
"²"	*yyout << "&sup2;";
"³"	*yyout << "&sup3;";
"å"	*yyout << "&aring;";
"ä"	*yyout << "&auml;";
"ö"	*yyout << "&ouml;";
"Å"	*yyout << "&Aring;";
"Ä"	*yyout << "&Auml;";
"Ö"	*yyout << "&Ouml;";

"™"	*yyout << "&tm;";
"µ"	*yyout << "&micro;";
"¶"	*yyout << "&para;";
"¬"	*yyout << "&not;";
"§"	*yyout << "&sect;";
"æ"	*yyout << "&sect;";
"¿"	*yyout << "&iquest;";
"¡"	*yyout << "&inexcl;";

{preprocessor}	echo(*yyout,"<code class = \"preprocessor\">", YYText(), "</code>");
{keyword}		echo(*yyout,"<code class = \"keyword\">", YYText(), "</code>");
{datatype}		echo(*yyout,"<code class = \"datatype\">", YYText(), "</code>");
{usertype}		echo(*yyout,"<code class = \"usertype\">", YYText(), "</code>");
{number}		echo(*yyout,"<code class = \"number\">", YYText(), "</code>");
{name}			*yyout << YYText();

"\\\""			*yyout << "\\&quot;";

"\"" {
	switch (mode)
	{
		case NORMAL_MODE:
			mode = STRING_MODE;
			*yyout << "<code class = \"string\">&quot;";
		break;
		case STRING_MODE:
			mode = NORMAL_MODE;
			*yyout << "&quot;</code>";
		break;
		default:
			*yyout << "&quot;";
	}
}

"'" {
	switch (mode)
	{
		case NORMAL_MODE:
			mode = CHAR_MODE;
			*yyout << "<code class = \"char\">'";
		break;
		case CHAR_MODE:
			mode = NORMAL_MODE;
			*yyout << "'</code>";
		break;
		default:
			*yyout << "'";
	}
}


%%

#include <iostream>
#include <string>
#include <fstream>

int main(int argc, char *argv[])
{
	if (argc < 3)
	{
		cerr << "usage: " << argv[0] << " destination_directory source_file1 source_file2 ... \n\n"
				"The output html-file will be printed to a file with the same name as the source file but with \".html\""
				" appended.\n";
		return 1;
	}

	string path = argv[1];
	if (path[path.length()-1] != '/') path += '/';

	while(argc > 2)
	{
		argc--;
		char *name = argv[argc];
		int i = strlen(name) - 1;
		while(i >= 0 && name[i] != '/') i--;

		if (name[i] == '/') name += i+1;

		string filename = path + name + ".html";

		ifstream in(argv[argc]);
		ofstream out(filename.c_str());

		cout << "generating: " << filename.c_str() << endl;
		out << "<html><head><meta name=\"generator\" contents=\"cpp2html by Arvid Norberg (arvid@iname.com) (this program can be found on http://www.bebits.com)\"><title>" << argv[argc] << "</title><link href=\"syntax.css\" rel=\"stylesheet\" type=\"text/css\"></head><body>\n<pre>";

		mode = NORMAL_MODE;
		FlexLexer* lexer = new yyFlexLexer(&in, &out);
		lexer->yylex();
		delete lexer;

		out << "</pre></body></html>";
	}z
	
	// print the syntax stylesheet file
	ofstream syntax((path + "syntax.css").c_str(), ios::noreplace | ios::out);
	syntax  <<	"body {background-color: #443b6e; color: #e6ba9d}\n"
				".number {color: yellow}\n"
				".keyword {color: white}\n"
				".preprocessor {color: white}\n"
				".datatype {color: white}\n"
				".usertype {color: grey}\n"
				".comment {color: #9190d1}\n"
				".string {color: #90ef90}\n"
				".char {color: yellow}\n";

	return 0;
}

void echo(ostream &out,const char *pre_string, const char *string, const char *post_string)
{
	if (mode == NORMAL_MODE)
		out << pre_string << string << post_string;
	else
		out << string;
}
